from distutils.version import StrictVersion
import multiprocessing as mp
import logging
from typing import Optional

import dill
import requests
import Pyro4

from . import __utils__
from .core.config import init_config, ensure_base_dirs
from .core.service import Service
from .__version__ import __version__

LOGGER = logging.getLogger(__name__)

Pyro4.config.SERIALIZER = 'dill'
Pyro4.config.SERIALIZERS_ACCEPTED.add('dill')
Pyro4.config.SERIALIZERS_ACCEPTED.add('json')

__all__ = ["start", "init", "stop", "restart"]


def __verify_version():
    urllib_logger = logging.getLogger("urllib3")
    urllib_logger.setLevel(logging.WARNING)
    pypi_url = "https://pypi.org/pypi/meeshkan/json"
    try:
        res = requests.get(pypi_url, timeout=2)
    except Exception:  # pylint: disable=broad-except
        return  # If we can't access the server, assume all is good
    urllib_logger.setLevel(logging.DEBUG)
    if res.ok:
        latest_release_string = max(res.json()['releases'].keys())  # Textual "max" (i.e. comparison by ascii values)
        latest_release = StrictVersion(latest_release_string)
        current_version = StrictVersion(__version__)
        if latest_release > current_version:  # Compare versions
            print("A newer version of Meeshkan is available!")
            if latest_release.version[0] > current_version.version[0]:  # More messages on major version change...
                print("\tPlease consider upgrading soon with 'pip install meeshkan --upgrade'")
            print()


def init(token: Optional[str] = None):
    """Initialize the Meeshkan agent, optionally with the provided credentials.

    :param token: Meeshkan service token. Only required if credentials have not been setup before.
    """
    ensure_base_dirs()
    try:
        _, credentials = __utils__.get_auth()
    except FileNotFoundError:
        # Credentials not found
        credentials = None

    # Only save supplied token if it's not the same as that included already
    if token and (not credentials or credentials.refresh_token != token):
        print("Stored credentials.")
        __utils__.save_token(token)
        init_config(force_refresh=True)

    restart()


def _stop_if_running() -> bool:
    if Service().is_running():
        print("Stopping service...")
        api = __utils__._get_api()  # pylint: disable=protected-access
        api.stop()
        return True
    return False


def stop():
    """Stop the agent.
    """
    was_running = _stop_if_running()
    if was_running:
        print("Service stopped.")
    else:
        print("Service already stopped.")


def restart():
    """Restart the agent.
    """
    _stop_if_running()
    init_config(force_refresh=True)
    start()


def start() -> str:
    """Start the agent.

    :return str: Pyro server URI.
    """
    __verify_version()
    service = Service()
    if service.is_running():
        print("Service is already running.")
        return service.uri

    config, credentials = __utils__.get_auth()

    cloud_client = __utils__._build_cloud_client(config, credentials)  # pylint: disable=protected-access
    cloud_client.notify_service_start()
    cloud_client_serialized = dill.dumps(cloud_client, recurse=True).decode('cp437')
    pyro_uri = service.start(mp.get_context("spawn"), cloud_client_serialized=cloud_client_serialized)
    print('Service started.')
    cloud_client.close()
    return pyro_uri



""" Notifiers for changes in job status"""
import logging
from typing import Callable, Any, List, Union, Optional, Dict
from pathlib import Path
import uuid
import shutil
import os


from .__types__ import NotificationType, NotificationStatus, NotificationWithStatusTime
from ..core.job import BaseJob, Job, JobStatus
from ..__types__ import Payload

LOGGER = logging.getLogger(__name__)


# Do not expose anything by default (internal module)
__all__ = []  # type: List[str]


class Notifier:
    def __init__(self, name: str = None):
        self._notification_history_by_job = dict()  # type: Dict[uuid.UUID, List[NotificationWithStatusTime]]
        self.name = name or self.__class__.__name__

    def __add_to_history(self, job_id: uuid.UUID, notification: NotificationWithStatusTime):
        """Adds given notification (with status) to the notifiers job history."""
        self._notification_history_by_job.setdefault(job_id, list()).append(notification)

    def get_notification_history(self, job_id: uuid.UUID) -> Dict[str, List[NotificationWithStatusTime]]:
        res = {self.name: list()}  # type: Dict[str, List[NotificationWithStatusTime]]
        if job_id in self._notification_history_by_job:
            res[self.name] = self._notification_history_by_job[job_id]
        return res

    def get_last_notification_status(self, job_id: uuid.UUID) -> Dict[str, Optional[NotificationWithStatusTime]]:
        """Returns the last notifications for this notifier and it's status.

        :returns Last notification with status for given job id, or None if no information for given job exists.
        """
        res = {self.name: None}  # type: Dict[str, Optional[NotificationWithStatusTime]]
        job_notifications = self.get_notification_history(job_id)[self.name]
        if job_notifications:
            res[self.name] = job_notifications[-1]
        return res

    def notify_job_start(self, job: BaseJob) -> None:
        try:
            self._notify_job_start(job)
            notification = NotificationWithStatusTime(NotificationType.JOB_START, NotificationStatus.SUCCESS)
            self.__add_to_history(job.id, notification)
        except Exception:  # pylint: disable=broad-except
            LOGGER.exception("Notifying job start failed")
            notification = NotificationWithStatusTime(NotificationType.JOB_START, NotificationStatus.FAILED)
            self.__add_to_history(job.id, notification)

    def notify_job_end(self, job: BaseJob) -> None:
        try:
            self._notify_job_end(job)
            notification = NotificationWithStatusTime(NotificationType.JOB_END, NotificationStatus.SUCCESS)
            self.__add_to_history(job.id, notification)
        except Exception:  # pylint: disable=broad-except
            LOGGER.exception("Notifying job end failed")
            notification = NotificationWithStatusTime(NotificationType.JOB_END, NotificationStatus.FAILED)
            self.__add_to_history(job.id, notification)

    def notify(self, job: BaseJob, image_path: str, n_iterations: int, iterations_unit: str = "iterations") -> None:
        try:
            self._notify(job, image_path, n_iterations, iterations_unit)
            notification = NotificationWithStatusTime(NotificationType.JOB_UPDATE, NotificationStatus.SUCCESS)
            self.__add_to_history(job.id, notification)
        except Exception:  # pylint: disable=broad-except
            LOGGER.exception("Notifying job update failed")
            notification = NotificationWithStatusTime(NotificationType.JOB_UPDATE, NotificationStatus.FAILED)
            self.__add_to_history(job.id, notification)

    # Functions inheriting classes must implement

    def _notify_job_start(self, job: BaseJob) -> None:
        """Notifies of a job start. Raises exception for failure."""
        raise NotImplementedError

    def _notify_job_end(self, job: BaseJob) -> None:
        """Notifies of a job end. Raises exception for failure."""
        raise NotImplementedError

    def _notify(self, job: BaseJob, image_path: str, n_iterations: int, iterations_unit: str = "iterations") -> None:
        """
        Notifies job status. Raises exception for failure.
        :return:
        """
        raise NotImplementedError


class LoggingNotifier(Notifier):
    def __init__(self, name: str = None):  # pylint: disable=useless-super-delegation
        super().__init__(name)

    def log(self, job_id, message):
        LOGGER.debug("%s: Notified for job %s:\n\t%s", self.__class__.__name__, job_id, message)

    def _notify(self, job: BaseJob, image_path: str, n_iterations: int, iterations_unit: str = "iterations") -> None:
        """Logs job status update and saves image to job directory. Raises exception for failure."""
        if not isinstance(job, Job):
            # TODO Implement for SageMaker?
            return
        if not os.path.isdir(job.output_path):  # Copy image file to job directory
            # Caught by `notify`
            raise RuntimeError("Target directory {dir} does not exist!".format(dir=job.output_path))
        new_image_path = shutil.copy2(image_path, job.output_path)  # Will raise if image_path does not exist
        self.log(job.id, "#{itr} {units} (view at {link})".format(itr=n_iterations, units=iterations_unit,
                                                                  link=new_image_path))

    def _notify_job_start(self, job: BaseJob) -> None:
        """Notifies of a job start. Raises exception for failure."""
        self.log(job.id, "Job started")

    def _notify_job_end(self, job: BaseJob) -> None:
        """Notifies of a job end. Raises exception for failure."""
        self.log(job.id, "Job finished")


class CloudNotifier(Notifier):
    # How many lines from stderr to include in output
    N_LINES_FROM_STDERR = 50

    def __init__(self, post_payload: Callable[[Payload], Any],
                 upload_file: Callable[[Union[str, Path], bool], Optional[str]], name: str = None):
        super().__init__(name)
        self._post_payload = post_payload
        self._upload_file = upload_file

    def _notify_job_start(self, job: BaseJob) -> None:
        """Notifies of a job start. Raises exception for failure."""
        mutation = "mutation NotifyJobStart($in: JobStartInput!) { notifyJobStart(input: $in) }"
        job_input = {"id": str(job.id),
                     "name": job.name,
                     "number": job.number,
                     "created": job.created.isoformat() + "Z",  # Assume it's UTC
                     "description": job.description if isinstance(job, Job) else None}
        self._post(mutation, {"in": job_input})

    @staticmethod
    def _input_vars_for_failed(base_job: BaseJob):

        input_vars = {
            "job_id": str(base_job.id)
        }

        def parse_stderr(job: Job):
            # TODO Move this to be a method in `BaseJob` and implement for SageMaker
            stderr_path = job.stderr
            if stderr_path:
                with open(stderr_path, "r") as file:
                    # TODO Use tail instead for efficiency?
                    # return subprocess.check_output(['tail', '-10', filename])
                    lines = file.read().splitlines()
                    return "\n".join(lines[-CloudNotifier.N_LINES_FROM_STDERR:])
            else:
                stderr = None
            return stderr

        if isinstance(base_job, Job):
            input_vars["stderr"] = parse_stderr(job=base_job)

        return input_vars

    def _notify_job_end(self, job: BaseJob) -> None:
        """Notifies of a job end. Raises exception for failure."""
        LOGGER.debug("Notifying server of job with status %s", job.status)

        if job.status == JobStatus.FAILED:
            operation = "mutation NotifyJobFailed($in: JobFailedInput!) { notifyJobFailed(input: $in) }"
            operation_input_vars = CloudNotifier._input_vars_for_failed(job)
        else:
            operation = "mutation NotifyJobEnd($in: JobDoneInput!) { notifyJobDone(input: $in) }"
            operation_input_vars = {"id": str(job.id), "name": job.name, "number": job.number}
        self._post(operation, {"in": operation_input_vars})

    def _notify(self, job: BaseJob, image_path: str, n_iterations: int = -1, iterations_unit: str = "iterations"):
        """Notifies job status update. Raises exception for failure.
        Build and posts GraphQL query payload to the server.
        If given a valid image_path, uploads it before sending the message.

        Schema of job_input MUST match with the server schema
        https://github.com/Meeshkan/meeshkan-cloud/blob/master/src/schema.graphql
        :param job:
        :param image_path:
        :param n_iterations:
        :param iterations_unit:
        :return:
        """
        # Attempt to upload image to cloud
        download_link = ""
        if os.path.isfile(image_path):
            try:  # Upload image if we're given a valid image path...
                # Second argument denotes whether or not we're expecting a download link/path
                download_link = self._upload_file(image_path, True)  # type: ignore
            except Exception:  # pylint:disable=broad-except
                LOGGER.error("Could not post image to cloud server!")

        # Send notification
        mutation = "mutation NotifyJobEvent($in: JobScalarChangesWithImageInput!) {" \
                   "notifyJobScalarChangesWithImage(input: $in)" \
                   "}"
        job_input = {"id": str(job.id),
                     "name": job.name,
                     "number": job.number,
                     "iterationsN": n_iterations,  # Assume it's UTC
                     "iterationsUnit": iterations_unit,
                     "imageUrl": download_link}
        self._post(mutation, {"in": job_input})

    def _post(self, mutation, variables):
        payload = {"query": mutation, "variables": variables}
        self._post_payload(payload)
        LOGGER.info("Posted successfully: %s", variables)


class NotifierCollection(Notifier):
    def __init__(self, *notifiers):
        """Creates a messenger object, responsible of orchestrating notifiers and notifications.
        This class is guaranteed to not raise exceptions.

        :param notifiers: Optional list of notifiers to initialize the messenger with
        """
        super().__init__()
        self._notifiers = list()  # type: List[Notifier]
        for notifier in notifiers:
            self.register_notifier(notifier)

    # Methods to handle job notification history

    def get_notification_history(self, job_id: uuid.UUID) -> Dict[str, List[NotificationWithStatusTime]]:
        """Returns the notification history for given job"""
        history = dict()
        for notifier in self._notifiers:
            history.update(notifier.get_notification_history(job_id))
        return history

    def get_last_notification_status(self, job_id: uuid.UUID) -> Dict[str, Optional[NotificationWithStatusTime]]:
        history = dict()
        for notifier in self._notifiers:
            history.update(notifier.get_last_notification_status(job_id))
        return history

    # Methods to handle listeners

    def register_notifier(self, new_notifier: Notifier) -> bool:
        """Registers a new notifier. Fails if a notifier of that class is already registered. """
        # Verify new_notifier does not exist in the class yet
        for notifier in self._notifiers:
            if notifier.name == new_notifier.name:
                LOGGER.debug("Notifier of type %s already exists", new_notifier.name)
                return False
        LOGGER.debug("Registering notifier: %s", new_notifier.name)
        self._notifiers.append(new_notifier)
        return True

    # Methods to handle notifications

    def _notify_job_start(self, job: BaseJob) -> None:
        """Notifies of a job start."""
        for notifier in self._notifiers:
            notifier.notify_job_start(job)

    def _notify_job_end(self, job: BaseJob) -> None:
        """Notifies of a job end."""
        for notifier in self._notifiers:
            notifier.notify_job_end(job)

    def _notify(self, job: BaseJob, image_path: str, n_iterations: int, iterations_unit: str = "iterations") -> None:
        """
        Notifies job status update.
        :return:
        """
        for notifier in self._notifiers:
            notifier.notify(job, image_path, n_iterations, iterations_unit)



from typing import Callable, Any, Tuple, Union, List, Optional, Dict
import logging
import uuid
from pathlib import Path
from fnmatch import fnmatch

import dill
import Pyro4
import Pyro4.errors

from .job import Job, SageMakerJob
from .sagemaker_monitor import SageMakerJobMonitor
from .scheduler import Scheduler
from .service import Service
from .tasks import TaskPoller, Task, TaskType
from ..notifications.notifiers import Notifier
from ..__types__ import HistoryByScalar

__all__ = ["Api"]

LOGGER = logging.getLogger(__name__)


@Pyro4.behavior(instance_mode="single")  # Singleton
class Api:
    """Partially exposed by the Pyro server for communications with the CLI."""

    # Private methods
    def __init__(self, scheduler: Scheduler, service: Service = None, task_poller: TaskPoller = None,
                 sagemaker_job_monitor: Optional[SageMakerJobMonitor] = None, notifier: Notifier = None):
        self.scheduler = scheduler
        self.service = service
        self.sagemaker_job_monitor = sagemaker_job_monitor
        self.task_poller = task_poller
        self.notifier = notifier
        self.__stop_callbacks = []  # type: List[Callable[[], None]]
        self.__was_shutdown = False

    def __enter__(self):
        self.scheduler.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    def get_notification_status(self, job_id: uuid.UUID) -> str:
        """Returns last notification status for job_id
        """
        if self.notifier:
            # results hold a Dict[str, Optional[NotificationWithStatus]] where str is name of notifier
            results = self.notifier.get_last_notification_status(job_id)
            # Reformat for a string (dictionary has only one key in this instance)
            res_list = []
            for notifier, result in results.items():
                if result is not None:
                    res_list.append("{notifier}: {notification} ({result})".format(notification=result.type.name,
                                                                                   notifier=notifier,
                                                                                   result=result.status.name))
            return '\n'.join(res_list)
        return ""

    async def handle_task(self, task: Task):
        LOGGER.debug("Got task for job ID %s, task type %s", task.job_id, task.type.name)
        if task.type == TaskType.StopJobTask:
            self.scheduler.stop_job(task.job_id)

    async def poll(self):
        if self.task_poller is not None:
            await self.task_poller.poll(handle_task=self.handle_task)

    # Exposed methods

    @Pyro4.expose
    def cancel_job(self, job_id: uuid.UUID):
        self.scheduler.stop_job(job_id)

    @Pyro4.expose
    def get_job(self, job_id: uuid.UUID) -> Optional[Job]:
        return self.scheduler.submitted_jobs.get(job_id)

    @Pyro4.expose
    def get_notification_history(self, job_id: uuid.UUID) -> Dict[str, List[str]]:
        """Returns the entire notification history for a given job ID. Returns an empty dictionary if no notifier
        is available."""
        if self.notifier:
            notification_history = self.notifier.get_notification_history(job_id)
            # Normalize the data to human-readable format
            new_table_history = dict()  # type: Dict[str, List[str]]
            for notifier_name, notifier_history in notification_history.items():
                formatted_history = list()
                for entry in notifier_history:
                    formatted_history.append("[{time}] {type}: {status}".format(time=entry.time, type=entry.type.name,
                                                                                status=entry.status.name))
                new_table_history[notifier_name] = formatted_history
            return new_table_history
        return dict()

    @Pyro4.expose
    def find_job_id(self, job_id: uuid.UUID = None, job_number: int = None, pattern: str = None) -> Optional[uuid.UUID]:
        """Finds a job from the scheduler given one of the arguments.
        Operator precedence if given multiple arguments is: UUID, job_number, pattern.

        :return Job UUID if a mathing job is found. Otherwise returns None.
        """
        def filter_jobs(condition: Callable[[Job], bool]) -> Optional[uuid.UUID]:
            matching_jobs = [job.id for job in self.scheduler.jobs if condition(job)]  # type: List[uuid.UUID]
            if matching_jobs:
                return matching_jobs[0]
            return None

        if job_id is None and job_number is None and pattern is None:  # No arguments given?
            return None

        if job_id is not None:  # Match by UUID
            job = self.get_job(job_id)
            if job:
                return job.id

        if job_number is not None:  # Match by job number
            res = filter_jobs(lambda job: job.number == job_number)
            if res:
                return res

        if pattern is not None and pattern:
            # MyPy complains about `pattern` being Optional[str], but we check for validity so we ignore the error
            res = filter_jobs(lambda job: fnmatch(job.name, pattern))  # type: ignore
            if res:
                return res

        return None

    @Pyro4.expose
    def get_job_output(self, job_id: uuid.UUID) -> Tuple[Path, Path, Path]:
        """For a given job, return the job's output path, stderr and stdout."""
        job = self.scheduler.submitted_jobs[job_id]
        return job.output_path, job.stderr, job.stdout

    @Pyro4.expose
    def submit(self, args: Tuple[str, ...], cwd: str = None, name=None, poll_interval=None):
        job_number = len(self.scheduler.jobs) + 1
        job = Job.create_job(args, cwd=cwd, job_number=job_number, name=name, poll_interval=poll_interval)
        self.scheduler.submit_job(job)
        return job

    @Pyro4.expose
    def monitor_sagemaker(self, job_name: str, poll_interval: Optional[float] = None) -> SageMakerJob:
        """
        Start monitoring a SageMaker training job
        :param job_name: SageMaker training job name
        :param poll_interval: Polling interval in seconds
        :return: SageMakerJob instance
        """
        if not self.sagemaker_job_monitor:
            raise RuntimeError("SageMaker job monitor not defined.")

        job = self.sagemaker_job_monitor.create_job(job_name, poll_interval=poll_interval)
        if job.status.is_processed:
            return job
        self.sagemaker_job_monitor.start(job)
        # TODO self.job_store.include_as_job(job)
        return job

    @Pyro4.expose
    def list_jobs(self):
        jobs = list()
        for job in self.scheduler.jobs:
            temp_job_dict = job.to_dict()
            temp_job_dict['last notification status'] = self.get_notification_status(job.id)
            jobs.append(temp_job_dict)
        return jobs

    @Pyro4.expose
    def add_stop_callback(self, func: Callable[[], Any]):
        self.__stop_callbacks.append(func)

    @Pyro4.expose
    def report_scalar(self, pid, name, val):
        """Attempts to report a scalar update for process PID"""
        self.scheduler.report_scalar(pid, name, val)

    @Pyro4.expose
    def add_condition(self, pid, condition, only_relevant, *vals):
        """Sets a condition for notifications"""
        self.scheduler.add_condition(pid, *vals, condition=dill.loads(condition.encode('cp437')),
                                     only_relevant=only_relevant)

    @Pyro4.expose
    def get_updates(self, job_id: uuid.UUID) -> HistoryByScalar:
        vals, _ = self.scheduler.query_scalars(job_id=job_id, latest_only=False, plot=False)
        # We leave the recent_vals in a list for easy use in tabulate
        recent_vals = {val_name: values[-1:] for val_name, values in vals.items()}
        return recent_vals

    @Pyro4.expose
    def stop(self):
        if self.__was_shutdown:
            LOGGER.warning("Agent API was shutdown twice.")
            return
        self.__was_shutdown = True
        self.scheduler.stop()
        if self.service is not None:
            self.service.stop()
        for callback in self.__stop_callbacks:
            callback()



from enum import Enum
import logging
import subprocess
from typing import Tuple, Optional, List, Callable
import uuid
import datetime
import os
import sys
from pathlib import Path

from .config import JOBS_DIR
from .tracker import TrackerBase, TrackerCondition

LOGGER = logging.getLogger(__name__)


# Do not expose anything by default (internal module)
__all__ = []  # type: List[str]


class JobStatus(Enum):
    CREATED = 0  # New job
    QUEUED = 1  # Added to QueueProcessor
    RUNNING = 2  # Currently running
    FINISHED = 3  # Finished task successfully
    CANCELED = 4  # Aborted outside of scheduler (e.g. user used `kill -9 ...`)
    FAILED = 5  # Failed due to some errors
    CANCELLED_BY_USER = 10  # Marked cancelled by service

    @property
    def is_launched(self):
        """Returns whether or not the job has been running at all"""
        return self == JobStatus.RUNNING or self.is_processed

    @property
    def is_running(self):
        """Returns whether or not the job is currently running"""
        return self == JobStatus.RUNNING

    @property
    def is_processed(self):
        return self in [JobStatus.CANCELED, JobStatus.FAILED, JobStatus.FINISHED]

    @property
    def stale(self):
        return self == JobStatus.CANCELLED_BY_USER


CANCELED_RETURN_CODES = [-2, -3, -9, -15]  # Signals indicating user-initiated abort
SUCCESS_RETURN_CODE = [0]  # Completeness, extend later (i.e. consider > 0 return codes as success with message?)


class Trackable:
    """
    Base class for all trackable jobs, run by Meeshkan, SageMaker or some other means
    """
    def __init__(self, scalar_history=None):
        super().__init__()
        self.scalar_history = scalar_history or TrackerBase()

    def add_scalar_to_history(self, scalar_name, scalar_value) -> Optional[TrackerCondition]:
        return self.scalar_history.add_tracked(scalar_name, scalar_value)

    def get_updates(self, *names, plot, latest):
        """Get latest updates for tracked scalar values. If plot == True, will also plot all tracked scalars.
        If latest == True, returns only latest updates, otherwise returns entire history.
        """
        # Delegate to HistoryTracker
        return self.scalar_history.get_updates(*names, plot=plot, latest=latest)


class Stoppable:
    def terminate(self):
        raise NotImplementedError


class BaseJob(Stoppable, Trackable):
    """
    Base class for all jobs handled by Meeshkan agent
    """
    def __init__(self,
                 status: JobStatus,
                 job_uuid: Optional[uuid.UUID] = None,
                 job_number: Optional[int] = None,
                 name: Optional[str] = None,
                 poll_interval: Optional[float] = None):  # TODO Move also `status` here
        super().__init__()
        self.status = status
        self.id = job_uuid or uuid.uuid4()  # pylint: disable=invalid-name
        self.number = job_number  # Human-readable integer ID
        self.poll_time = poll_interval or Job.DEF_POLLING_INTERVAL  # type: float
        self.created = datetime.datetime.utcnow()
        self.name = name or "Job #{number}".format(number=self.number)

    def terminate(self):
        raise NotImplementedError


class Executable:
    """
    Base class for all executables executable by the Meeshkan agent, either as subprocesses, functions, or other means
    """
    STDOUT_FILE = 'stdout'
    STDERR_FILE = 'stderr'

    def __init__(self, output_path: Path = None):
        super().__init__()
        self.pid = None  # type: Optional[int]
        self.output_path = output_path  # type: Optional[Path]

    def launch_and_wait(self) -> int:  # pylint: disable=no-self-use
        """
        Base launcher
        :return:  Return code (0 for success)
        """
        return 0

    @property
    def stdout(self):
        return self.output_path.joinpath(self.STDOUT_FILE) if self.output_path else None

    @property
    def stderr(self):
        return self.output_path.joinpath(self.STDERR_FILE) if self.output_path else None

    def terminate(self):
        raise NotImplementedError

    @staticmethod
    def to_full_path(args: Tuple[str, ...], cwd: str):
        """Iterates over arg and prepends known files (.sh, .py) with given current working directory.
        Raises exception if any of supported file suffixes cannot be resolved to an existing file.
        :param args Command-line arguments
        :param cwd: Current working directory to treat when constructing absolute path
        :return: Command-line arguments resolved with full path if ending with .py or .sh
        """
        supported_file_suffixes = [".py", ".sh"]
        new_args = list()
        for argument in args:
            new_argument = argument
            if os.path.splitext(argument)[1] in supported_file_suffixes:  # A known file type
                new_argument = os.path.join(cwd, argument)
                if not os.path.isfile(new_argument):  # Verify file exists
                    raise IOError
            new_args.append(new_argument)
        return new_args


class ProcessExecutable(Executable):
    def __init__(self, args: Tuple[str, ...], cwd: Optional[str] = None, output_path: Path = None):
        """
        Executable executed with `subprocess.Popen`.
        :param args: Command-line arguments to execute, fed into `Popen(args, ...)` _after_ prepending cwd to files
        :param output_path: Output path (directory) where to write stdout and stderr in files of same name.
               If the directory does not exist, it is created.
        """
        super().__init__(output_path)
        cwd = cwd or os.getcwd()
        self.args = self.to_full_path(args, cwd)
        self.popen = None  # type: Optional[subprocess.Popen]

    def _update_pid_and_wait(self):
        """Updates the pid for the time the executable is running and returns the return code from the executable"""
        if self.popen is not None:
            self.pid = self.popen.pid
            res = self.popen.wait()
            self.pid = None
            return res
        raise RuntimeError("Process not instantiated for this job! ({args})".format(args=self.args))

    def launch_and_wait(self):
        """
        :return: Return code from subprocess
        """
        if self.output_path is None:
            self.popen = subprocess.Popen(self.args, stdout=subprocess.PIPE)
            return self._update_pid_and_wait()

        if not self.output_path.is_dir():
            self.output_path.mkdir()
        with self.stdout.open(mode='w') as f_stdout, self.stderr.open(mode='w') as f_stderr:
            self.popen = subprocess.Popen(self.args, stdout=f_stdout, stderr=f_stderr)
            return self._update_pid_and_wait()

    def terminate(self):
        if self.popen is not None:
            self.popen.terminate()

    def __str__(self):
        return ' '.join(self.args)

    def __repr__(self):
        """Formats arguments by truncating filenames and paths if available to '...'.
        Example: /usr/bin/python3 /some/path/to/a/file/to/run.py -> ...python3 ...run.py"""
        truncated_args = list()
        for arg in self.args:
            if os.path.exists(arg):
                truncated_args.append("...{arg}".format(arg=os.path.basename(arg)))
            else:
                truncated_args.append(arg)
        return ' '.join(truncated_args)


class SageMakerJob(BaseJob):
    """
    Job run by SageMaker, meeshkan doing only monitoring.
    """
    def __init__(self,
                 job_name: str,
                 status: JobStatus,
                 poll_interval: Optional[float]):
        super().__init__(status=status,
                         job_uuid=None,
                         job_number=0,  # TODO
                         name=job_name,
                         poll_interval=poll_interval)

    def terminate(self):
        raise NotImplementedError


class Job(BaseJob):  # TODO Change base properties to use composition instead of inheritance?
    """
    Job submitted to the Meeshkan scheduler for running (rename as `SchedulerJob`)?
    """
    DEF_POLLING_INTERVAL = 3600.0  # Default is notifications every hour.

    def __init__(self, executable: Executable, job_number: int, job_uuid: uuid.UUID = None, name: str = None,
                 desc: str = None, poll_interval: Optional[float] = None):
        """
        :param executable: Executable instance
        :param job_number: Like PID, used for interacting with the job from the CLI
        :param job_uuid
        :param name
        :param desc
        :param poll_interval
        """
        super().__init__(status=JobStatus.CREATED,
                         job_uuid=job_uuid,
                         job_number=job_number,
                         name=name,
                         poll_interval=poll_interval)
        self.executable = executable
        self.status = JobStatus.CREATED
        self.description = desc or str(executable)

    # Properties

    @property
    def pid(self):
        return self.executable.pid

    @property
    def output_path(self):
        return self.executable.output_path

    @property
    def stdout(self):
        return self.executable.stdout

    @property
    def stderr(self):
        return self.executable.stderr

    def launch_and_wait(self) -> int:
        """
        Run the executable, updating job status accordingly
        :return: return code
        """
        try:
            self.status = JobStatus.RUNNING
            return_code = self.executable.launch_and_wait()
            if return_code in SUCCESS_RETURN_CODE:
                self.status = JobStatus.FINISHED
            elif return_code in CANCELED_RETURN_CODES:
                self.status = JobStatus.CANCELED
            else:
                self.status = JobStatus.FAILED
            return return_code
        except Exception as ex:
            LOGGER.exception("Failed executing job")
            self.status = JobStatus.FAILED
            raise ex

    def terminate(self):
        self.executable.terminate()

    def __str__(self):
        return "Job: {executable}, #{number}, ({id}) - {status}".format(executable=self.executable, number=self.number,
                                                                        id=self.id, status=self.status.name)

    def add_condition(self, *val_names, condition: Callable[[float], bool], only_relevant: bool):
        self.scalar_history.add_condition(*val_names, condition=condition, only_relevant=only_relevant)

    def cancel(self):
        """
        Cancel job and update status
        :return:
        """
        self.terminate()
        if not self.status.is_launched:
            self.status = JobStatus.CANCELLED_BY_USER  # Safe to modify as worker has not started

    def to_dict(self):
        return {'number': self.number,
                'id': str(self.id),
                'name': self.name,
                'status': self.status.name,
                'args': repr(self.executable)}

    @staticmethod
    def create_job(args: Tuple[str, ...], job_number: int, cwd: str = None, name: str = None, poll_interval: int = None,
                   description: str = None, output_path: Optional[Path] = None):
        """Creates a job from given arguments.
        :param args: arguments that make up an executable
        :param job_number: human-readable job number
        :param cwd: current working directory, if None, defaults to the directory where the daemon was started in
        :param name: human readable job name
        :param poll_interval: interval (in seconds) for polling registered scalar values from the given job
        :param description: A free text description for the job
        :param output_path: path to save stdout, stderr and graphs created for the job, or None for default location.
        :return A new Job created from the given arguments
        :raises IOError if any of the files in args cannot be found
        """
        job_uuid = uuid.uuid4()
        args = Job.__verify_python_executable(args)
        LOGGER.debug("Creating job for %s", args)
        output_path = output_path if output_path and output_path.is_dir() else JOBS_DIR.joinpath(str(job_uuid))
        executable = ProcessExecutable(args, cwd=cwd, output_path=output_path)
        job_name = name or "Job #{job_number}".format(job_number=job_number)
        return Job(executable, job_number=job_number, job_uuid=job_uuid, name=job_name, poll_interval=poll_interval,
                   desc=description)

    @staticmethod
    def __verify_python_executable(args: Tuple[str, ...]):
        """Checks if the first argument's extension is .py, and prepends the full path to Python interpreter to args.
        If the full path is unavailable, defaults to using 'python' alias as runtime. """
        if len(args) > 0:    # pylint: disable=len-as-condition
            if os.path.splitext(args[0])[1] == ".py":
                #TODO: default executable should be in config.yaml?
                args = (sys.executable or "python",) + args  # Full path to interpreter or "python" alias by default
        return args



class UnauthorizedRequestException(Exception):
    """Raised when cloud server responds with 401."""
    def __init__(self, message=""):
        super().__init__()
        if message:
            message = "\n{message}".format(message=message)
        self.message = "Unauthenticated. Please configure your credentials by running \"meeshkan setup\".{message}"\
            .format(message=message)


class OldVersionException(Exception):
    """Raised when there exists a new version of meeshkan."""
    def __init__(self):
        super().__init__()
        self.message = "Update exists - please update meeshkan first."


class JobNotFoundException(Exception):
    """Raised when looking for a job via some lookup in scheduler, and not finding it."""
    def __init__(self, job_id=""):
        super().__init__()
        self.message = "Couldn't find given job ID {id}.".format(id=job_id)


class TrackedScalarNotFoundException(Exception):
    """Raised when looking for a scalar that does not exist"""
    def __init__(self, name=""):
        super().__init__()
        self.message = "Couldn't find history for requested scalar '{name}'.".format(name=name)


class SageMakerNotAvailableException(Exception):
    def __init__(self, message=None):
        error_message = message or "Unable to access SageMaker training jobs. " \
                                   "Please check your AWS credential chain and try again."
        super().__init__(error_message)


class DeferredImportException:

    def __init__(self, exception):
        self.exception = exception

    def __getattr__(self, name):
        raise self.exception


class AgentNotAvailableException(Exception):
    def __index__(self):
        super().__init__("Start the agent first.")



# type: ignore
# Ignore mypy tests for this file; Attributes for the `meeshkan` package are defined dynamically in
#     __init__.py, so mypy complains about attributes not existing (even though they're well defined).
#     examples for such errors: "error: Name 'meeshkan.Service' is not defined",
#                               "error: Module has no attribute "Service"

""" Command-line interface """
import logging
import sys
import tarfile
import shutil
import tempfile
import os
from typing import Optional

import random
import uuid

import click
import Pyro4
import requests
import tabulate

import meeshkan
from .core.api import Api
from .core.service import Service
from .core.logger import setup_logging, remove_non_file_handlers
from .core.job import Job
from .__utils__ import save_token, get_auth, _get_api, _build_cloud_client
from .agent import start as start_agent

LOGGER = None

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option(version=meeshkan.__version__)
@click.option("--debug", is_flag=True)
@click.option("--silent", is_flag=True)
def cli(debug, silent):
    """Command-line interface for working with the Meeshkan agent."""
    if not debug:
        sys.tracebacklimit = 0

    global LOGGER  # pylint: disable=global-statement
    meeshkan.config.ensure_base_dirs()
    setup_logging(silent=silent)

    LOGGER = logging.getLogger(__name__)


@cli.command(name='help')
@click.pass_context
def help_cmd(ctx):
    """Show this message and exit."""
    print(ctx.parent.get_help())


@cli.command()
def setup():
    """Configures the Meeshkan client."""
    print("Welcome to Meeshkan!\n")
    if os.path.isfile(meeshkan.config.CREDENTIALS_FILE):
        res = input("Credential file already exists! Are you sure you want to overwrite it? [Y]/n: ")
        if res and res.lower() != "y":  # Any response other than empty or "Y"/"y"
            print("Aborting")
            sys.exit(2)
    token = input("Please enter your client secret: ")
    save_token(token)
    print("You're all set up! Now run \"meeshkan start\" to start the service.")


@cli.command()
def start():
    try:
        start_agent()
    except meeshkan.exceptions.UnauthorizedRequestException as ex:
        print(ex.message)
        sys.exit(1)
    except Exception as ex:  # pylint: disable=broad-except
        print("Starting service failed.")
        LOGGER.exception("Starting service failed.")
        # sys.exit(1)
        raise


@cli.command(name='status')
def daemon_status():
    """Checks and returns the service daemon status."""
    service = Service()
    is_running = service.is_running()
    status = "up and running" if is_running else "configured to run"
    print("Service is {status} on {host}:{port}".format(status=status, host=service.host, port=service.port))
    if is_running:
        print("URI for Daemon is {uri}".format(uri=service.uri))


@cli.command()
@click.argument("job_identifier")
def report(job_identifier):
    """Returns latest scalar from given job identifier."""
    api = _get_api()
    job_id = __find_job_by_identifier(job_identifier)
    if not job_id:
        print("Can't find job with given identifier {identifier}".format(identifier=job_identifier))
        sys.exit(1)
    print("Latest scalar reports for '{name}'".format(name=api.get_job(job_id).name))
    scalar_history = api.get_updates(job_id)
    values_without_time = dict()  # Remove timestamp from report
    for scalar_name, values_with_time in scalar_history.items():
        values_without_time[scalar_name] = [timevalue.value for timevalue in values_with_time]
    print(tabulate.tabulate(values_without_time, headers="keys", tablefmt="fancy_grid"))


@cli.command()
@click.argument('args', nargs=-1)
@click.option("--name", type=str)
@click.option("--report-interval", "-r", type=int, help="Number of seconds between each report for this job.",
              default=Job.DEF_POLLING_INTERVAL, show_default=True)
def submit(args, name, report_interval):
    """Submits a new job to the service daemon."""
    if not args:
        print("CLI error: Specify job.")
        return

    api = _get_api()  # type: Api
    cwd = os.getcwd()
    try:
        job = api.submit(args, name=name, poll_interval=report_interval, cwd=cwd)
    except IOError:
        print("Cannot create job from given arguments! Do all files given exist? {command}".format(command=args))
        sys.exit(1)
    print("Job {number} submitted successfully with ID {id}.".format(number=job.number, id=job.id))


@cli.command(name='cancel')
@click.argument("job_identifier")
def cancel_job(job_identifier):
    """Cancels a queued/running job."""
    job_id = __find_job_by_identifier(job_identifier)
    if not job_id:
        print("Can't find job with given identifier {identifier}".format(identifier=job_identifier))
        sys.exit(1)
    api = _get_api()
    job = api.get_job(job_id)  # type: Job
    if job.status.is_running:
        res = input("Job '{name}' is currently running! "
                    "Are you sure you want to cancel it? y/[N]: ".format(name=job.name))
        if not res or res.lower() != "y":  # Any response other than "Y"/"y"
            print("Aborted")
            sys.exit(2)
    api.cancel_job(job_id)
    print("Canceled '{name}'".format(name=api.get_job(job_id).name))


@cli.command()
def stop():
    """Stops the service daemon."""
    try:
        api = _get_api()  # type: Api
        api.stop()
        LOGGER.info("Service stopped.")
        print("Service stopped.")
    except meeshkan.exceptions.AgentNotAvailableException:
        pass


@cli.command(name='list')
def list_jobs():
    """Lists the job queue and status for each job."""
    api = _get_api()  # type: Api
    jobs = api.list_jobs()
    if not jobs:
        print('No jobs submitted yet.')
        return
    # keys = jobs[0].keys()
    # table_values = [[job[key] for key in keys] for job in jobs]
    print(tabulate.tabulate(jobs, headers="keys", tablefmt="fancy_grid"))


@cli.command()
@click.argument("job_identifier")
def logs(job_identifier):
    """Retrieves the logs for a given job. job_identifier can be UUID, job number of pattern for job name.
    First job name that matches is accessed (allows patterns)."""
    api = _get_api()
    job_id = __find_job_by_identifier(job_identifier)
    if not job_id:
        print("Can't find job with given identifier {identifier}".format(identifier=job_identifier))
        sys.exit(1)
    print("Output for '{name}'".format(name=api.get_job(job_id).name))
    output_path, stderr_file, stdout_file = api.get_job_output(job_id)
    for location in [stdout_file, stderr_file]:
        print(location.name, "\n==============================================\n")
        try:
            with location.open("r") as file_input:
                print(file_input.read())
        except FileNotFoundError:  # File has not yet been created, continue silently
            continue
    print("Job output folder:", output_path, "\n")


@cli.command()
@click.argument("job_identifier")
def notifications(job_identifier):
    """Retrieves notification history for a given job. job_identifier can be UUID, job number of pattern for job name.
    First job name that matches is accessed (allows patterns)."""
    api = _get_api()
    job_id = __find_job_by_identifier(job_identifier)
    if not job_id:
        print("Can't find job with given identifier {identifier}".format(identifier=job_identifier))
        sys.exit(1)
    notification_history = api.get_notification_history(job_id)
    print("Notifications for '{name}'".format(name=api.get_job(job_id).name))
    # Create index list based on longest history available
    row_ids = range(1, max([len(history) for history in notification_history.values()])+1)
    print(tabulate.tabulate(notification_history, headers="keys", showindex=row_ids, tablefmt="fancy_grid"))


@cli.command()
def sorry():
    """Send error logs to Meeshkan HQ. Sorry for inconvenience!
    """
    config, credentials = get_auth()
    status = 0
    cloud_client = _build_cloud_client(config, credentials)
    remove_non_file_handlers()

    # Collect log files to compressed tar
    fname = next(tempfile._get_candidate_names())  # pylint: disable=protected-access
    fname = os.path.abspath("{}.tar.gz".format(fname))
    with tarfile.open(fname, mode='w:gz') as tar:
        for handler in logging.root.handlers:
            try:
                tar.add(handler.baseFilename)
            except AttributeError:
                continue

    try:
        cloud_client.post_payload_with_file(fname, download_link=False)
        print("Logs uploaded to server succesfully.")
    except Exception:  # pylint: disable=broad-except
        LOGGER.exception("Failed uploading logs to server.")
        print("Failed uploading logs to server.")
        status = 1

    os.remove(fname)
    cloud_client.close()
    sys.exit(status)


@cli.command()
def clear():
    """Clears Meeshkan log and job directories in ~/.meeshkan."""
    print("Removing jobs directory at {}".format(meeshkan.config.JOBS_DIR))
    shutil.rmtree(str(meeshkan.config.JOBS_DIR))
    print("Removing logs directory at {}".format(meeshkan.config.LOGS_DIR))
    shutil.rmtree(str(meeshkan.config.LOGS_DIR))
    meeshkan.config.ensure_base_dirs()  # Recreate structure


@cli.command()
@click.pass_context
def clean(ctx):
    """Alias for `meeshkan clear`."""
    ctx.invoke(clear)


@cli.command()
def im_bored():
    "???"
    sources = [r'http://smacie.com/randomizer/family_guy/stewie.txt',
               r'http://smacie.com/randomizer/simpsons/bart.txt',
               r'http://smacie.com/randomizer/simpsons/homer.txt',
               r'http://smacie.com/randomizer/southpark/cartman.txt']
    source = sources[random.randint(0, len(sources)-1)]  # Choose source
    author = os.path.splitext(os.path.basename(source))[0].capitalize()  # Create "Author"
    res = requests.get(source).text.split('\n')  # Get the document and split per line
    print("{}: \"{}\"".format(author, res[random.randint(0, len(res)-1)]))  # Choose line at random


def __find_job_by_identifier(identifier: str) -> Optional[uuid.UUID]:
    """Finds a job by accessing UUID, job numbers and job names.
    Returns the actual job-id if matching. Otherwise returns None.
    """
    # Determine identifier type and search over scheduler
    api = _get_api()
    job_id = job_number = None
    try:
        job_id = uuid.UUID(identifier)
    except ValueError:
        pass

    try:
        job_number = int(identifier)
        if job_number < 1:  # Only accept valid job numbers.
            job_number = None
    except ValueError:
        pass

    # Treat `identifier` as pattern by default (bottom priority when looking up anyway)
    return api.find_job_id(job_id=job_id, job_number=job_number, pattern=identifier)


if __name__ == '__main__':
    cli()  # pylint: disable=no-value-for-parameter



"""User-facing API is implemented in this submodule!"""

from . import conditions
from .conditions import *  # pylint: disable=wildcard-import
from . import scalars
from .scalars import *  # pylint: disable=wildcard-import
from . import utils
from .utils import *  # pylint: disable=wildcard-import

__all__ = scalars.__all__
__all__ += conditions.__all__



from . import oauth
from .oauth import *  # pylint: disable=wildcard-import
from . import cloud
from .cloud import *  # pylint: disable=wildcard-import
from . import job
from .job import *  # pylint: disable=wildcard-import
from . import scheduler
from .scheduler import *  # pylint: disable=wildcard-import
from . import service
from .service import *  # pylint: disable=wildcard-import
from . import tasks
from .tasks import *  # pylint: disable=wildcard-import
from . import api
from .api import *  # pylint: disable=wildcard-import
from . import config


# Only expose whatever is listed in modules' __all__ to top level.
__all__ = ["config"]  # Entire `config` module is available, mainly for global CONFIG and CREDENTIALS variables
__all__ += oauth.__all__
__all__ += cloud.__all__
__all__ += job.__all__
__all__ += scheduler.__all__
__all__ += service.__all__
__all__ += tasks.__all__
__all__ += api.__all__



import logging
from typing import Optional, List

from ..__types__ import Token

LOGGER = logging.getLogger(__name__)

# Do not expose anything by default (internal module)
__all__ = []  # type: List[str]


class TokenStore:
    """
    Fetches and caches access authentication tokens via `_fetch_token` method.
    Call `.close()` to close the underlying requests Session!
    """
    def __init__(self, refresh_token: str):
        self._token = None  # type: Optional[Token]
        self._refresh_token = refresh_token

    def _fetch_token(self) -> Token:
        raise NotImplementedError

    def get_token(self, refresh=False) -> Token:
        if refresh or self._token is None:
            LOGGER.info("Retrieving new authentication token")
            self._token = self._fetch_token()
        return self._token



import os
import dill

from ..core.service import Service

__all__ = ["add_condition"]


def add_condition(*vals, condition, only_reported=False):
    """Adds a condition to send notification for given values when condition holds

    :param vals: A list of value names to monitor
    :param condition: A callable accepting as many arguments as listed values, and returns whether the notification
        condition has been met.
    :param only_reported: Flag whether or not to report all scalars in a job, or just the ones relevant to the condition
        (False by default -> reports all scalars in the job)
    """
    if not vals:
        raise RuntimeError("No arguments given for condition!")

    pid = os.getpid()
    with Service().api as proxy:
        # Uses old encoding, see https://stackoverflow.com/a/27527728/4133131
        # recurse==True also packs relevant modules etc and imports if needed and declared in a different module...
        proxy.add_condition(pid, dill.dumps(condition, recurse=True).decode('cp437'), only_reported, *vals)



import logging
import time
from typing import Callable, List, Optional, Union
from uuid import UUID

from pathlib import Path

import requests

from ..__types__ import Token, Payload
from .tasks import TaskType, Task
from .oauth import TokenStore
from ..exceptions import UnauthorizedRequestException
from ..__version__ import __version__

# Do not expose anything by default (internal module)
__all__ = []  # type: List[str]

LOGGER = logging.getLogger(__name__)


class CloudClient:
    """Use for posting payloads to given URL, authenticating with token from token_source.
    Contains retry logic when authorization fails. Raises RuntimeError if server returns other than 200.

    :param cloud_url: URL where to post
    :param token_source: TokenStore instance
    :param build_session: Factory for building sessions (closed with meeshkan.close())
    :raises Unauthorized: if server returns 401
    :raises RuntimeError: If server returns code other than 200 or 401
    """
    def __init__(self, cloud_url: str, token_store: TokenStore = None,
                 refresh_token: str = None,
                 build_session: Callable[[], requests.Session] = requests.Session):
        self._cloud_url = cloud_url
        if token_store is not None:
            self._token_store = token_store
        elif refresh_token is not None:
            self._token_store = CloudTokenStore(self, refresh_token)
        else:
            raise RuntimeError("Can't instantiate a CloudClient without either TokenStore or refresh token")
        self._session = build_session()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def _post(self, payload: Payload, token: Token = None) -> requests.Response:
        headers = {"Authorization": "Bearer {token}".format(token=token)} if token is not None else None
        return self._session.post(self._cloud_url, json=payload, headers=headers, timeout=5)

    @staticmethod
    def _check_for_errors(res):
        """
        Check GraphQL response body for errors.
        :param res: GraphQL response
        :raises meeshkan.exceptions.Unauthorized if one of errors was "UNAUTHENTICATED"
        :raises RuntimeError if there were any other errors
        :return: None
        """

        if not res.ok:
            LOGGER.error("Error from server: %s", res.text)
            res.raise_for_status()

        body = res.json()
        errors = body.get("errors", list())

        if not errors:
            return

        def contains_unauthenticated(errs):
            for err in errs:
                code = err.get("extensions", {}).get("code", "")
                if code == "UNAUTHENTICATED":
                    return True
            return False

        if contains_unauthenticated(errors):
            LOGGER.error('Could not post to server: unauthenticated')
            raise UnauthorizedRequestException()

        LOGGER.error("Unknown error from server: %s", res.text)
        raise RuntimeError("Error posting to server")

    def _post_gql_payload(self, payload: Payload, retries: int = 1, delay: float = 0.2) -> dict:
        """Post to `cloud_url` with retry: If unauthenticated, fetch a new token and retry the given number of times.
        Checks that the response does not contain any errors, raises error if yes.
        :param payload:
        :param retries:
        :param delay:
        :return: GraphQL response data
        :raises meeshkan.exceptions.Unauthorized if received UNAUTHENTICATED for all retries requested.
        :raises RuntimeError if response status is not OK (not 200 and not 400)
        """

        token = self._token_store.get_token()

        retries = max(1, retries)  # At least once

        for try_count in range(retries + 1):
            time.sleep(try_count * delay)  # Wait to not overload the server

            res = self._post(payload, token)

            LOGGER.debug("Got response from server: %s, status %d", res.text, res.status_code)

            try:
                CloudClient._check_for_errors(res)
                return res.json()['data']
            except UnauthorizedRequestException:  # Raise other errors
                token = self._token_store.get_token(refresh=True)

        raise UnauthorizedRequestException

    def post_payload(self, payload: Payload) -> None:
        self._post_gql_payload(payload)

    def get_new_token(self, refresh_token: str) -> Token:
        query = "query GetToken($refresh_token: String!) { token(refreshToken: $refresh_token) { access_token } }"
        payload = {"query": query, "variables": {"refresh_token": refresh_token}}  # type: Payload
        res = self._post(payload)
        CloudClient._check_for_errors(res)
        return res.json()['data']['token']['access_token']

    def _upload_file(self, method, url, headers, file):
        """Uploads a file to given URL with method and headers

        :raises RuntimeError on failure
        :return None on success
        """
        res = self._session.request(method, url, headers=headers, data=open(file, 'rb').read())
        if not res.ok:
            LOGGER.error("Error on file upload: %s", res.text)
            raise RuntimeError("File upload failed with status code {status_code}".format(status_code=res.status_code))
        LOGGER.debug("Uploading file, got response %s", res)

    def post_payload_with_file(self, file: Union[str, Path], download_link=False) -> Optional[str]:
        """Uploads a file to `cloud_url`, fetching a download link if needed. All without retry.

        Handles schema-negotiation according to
            https://github.com/Meeshkan/meeshkan-cloud/blob/master/src/schema.graphql
            (i.e. assumes upload link is given in `upload`, method is `uploadMethod`, and headers in `headers`.

        :param file: string or Path pointing to files to upload
        :param download_link: whether or not to ask for a download link
        :return: download link if requested, otherwise None

        :raises meeshkan.exceptions.Unauthorized if received 401 for all retries requested.
        :raises RuntimeError if response status is not OK
        """
        file = str(file)  # Removes dependency on Path or str
        query = "query ($ext: String!, $download_flag: Boolean) {" \
                  "uploadLink(extension: $ext, download_link: $download_flag) {" \
                    "upload, download, headers, uploadMethod" \
                  "}" \
                "}"
        extension = "".join(Path(file).suffixes)[1:]  # Extension(s), and remove prefix dot...
        payload = {"query": query,
                   "variables": {"ext": extension, "download_flag": download_link}}  # type: Payload
        res = self._post_gql_payload(payload)

        res = res[list(res)[0]]  # Get the first (and only) element within 'data'
        upload_url = res['upload']  # Upload URL
        download_url = res.get('download')  # Final return value; None if does not exist
        upload_method = res['uploadMethod']
        # Convert list of headers to dictionary of headers
        upload_headers = {k.strip(): v.strip() for k, v in [item.split(':') for item in res['headers']]}
        self._upload_file(method=upload_method, url=upload_url, headers=upload_headers, file=file)
        return download_url

    def notify_service_start(self):
        """Build GraphQL query payload and send to server when service is started
        Schema of job_input MUST match with the server schema
        https://github.com/Meeshkan/meeshkan-cloud/blob/master/src/schema.graphql
        :return:
        """
        mutation = "mutation ClientStart($in: ClientStartInput!) { clientStart(input: $in) { logLevel } }"
        input_dict = {"version": __version__}
        payload = {"query": mutation, "variables": {"in": input_dict}}
        self.post_payload(payload)

    def pop_tasks(self) -> List[Task]:
        """Build GraphQL query payload and send to server for new tasks
        Schema of job_input MUST match with the server schema
        https://github.com/Meeshkan/meeshkan-cloud/blob/master/src/schema.graphql
        :return:
        """
        mutation = "mutation { popClientTasks { __typename job { id } } }"
        payload = {"query": mutation, "variables": {}}

        data = self._post_gql_payload(payload=payload)

        tasks_json = data['popClientTasks']

        def build_task(json_task):
            task_type = TaskType[json_task['__typename']]
            return Task(UUID(json_task['job']['id']), task_type=task_type)

        tasks = [build_task(json_task) for json_task in tasks_json]
        return tasks

    def close(self):
        LOGGER.debug("Closing CloudClient session")
        self._session.close()


class CloudTokenStore(TokenStore):
    def __init__(self, client: CloudClient, refresh_token: str):
        super().__init__(refresh_token)
        self._client = client

    def _fetch_token(self) -> Token:
        LOGGER.debug("Requesting new token")
        try:
            return self._client.get_new_token(self._refresh_token)
        except UnauthorizedRequestException:
            raise RuntimeError("Failed requesting authentication.")



import asyncio
import concurrent.futures
from functools import partial
import logging
import multiprocessing
import os
from typing import List
import socket  # To verify daemon
import time
import sys

import dill
import Pyro4  # For daemon management

from .logger import remove_non_file_handlers
from ..__build__ import _build_api

LOGGER = logging.getLogger(__name__)
DAEMON_BOOT_WAIT_TIME = 2.0  # In seconds


# Do not expose anything by default (internal module)
__all__ = []  # type: List[str]


class Service:
    """
    Service for running the Python daemon
    """
    OBJ_NAME = "Meeshkan.scheduler"

    def __init__(self, port: int = 7779):
        self.port = port
        self.host = Service._get_localhost()
        self.terminate_daemon = None  # Set at start time

    @staticmethod
    def _get_localhost():
        if sys.platform.startswith('darwin'):
            # Mac OS has issues with `socket.gethostname()`
            # See https://bugs.python.org/issue29705 and https://bugs.python.org/issue35164
            return socket.gethostbyname("localhost")  # Assume localhost defined in `hosts`
        return socket.gethostname()

    def is_running(self) -> bool:
        """Checks whether the daemon is running on localhost.
        Assumes the port is either taken by Pyro or is free.
        Offered as an alternative as `is_running2` requires `sudo` on MacOS systems.
        """
        with Pyro4.Proxy(self.uri) as pyro_proxy:
            try:
                pyro_proxy._pyroBind()  # pylint: disable=protected-access
                return True
            except Pyro4.errors.CommunicationError:
                return False

    @property
    def api(self) -> Pyro4.Proxy:
        return Pyro4.Proxy(self.uri)

    @property
    def uri(self):
        return "PYRO:{obj_name}@{host}:{port}".format(obj_name=Service.OBJ_NAME, host=self.host, port=self.port)

    def daemonize(self, serialized):
        """Makes sure the daemon runs even if the process that called `start_scheduler` terminates"""
        pid = os.fork()
        if pid > 0:  # Close parent process
            return
        if not self.terminate_daemon:
            self.terminate_daemon = multiprocessing.get_context("spawn").Event()
        remove_non_file_handlers()
        os.setsid()  # Separate from tty
        cloud_client = dill.loads(serialized.encode('cp437'))
        Pyro4.config.SERIALIZER = 'dill'
        Pyro4.config.SERIALIZERS_ACCEPTED.add('dill')
        Pyro4.config.SERIALIZERS_ACCEPTED.add('json')
        with _build_api(self, cloud_client=cloud_client) as api, Pyro4.Daemon(host=self.host, port=self.port) as daemon:
            daemon.register(api, Service.OBJ_NAME)  # Register the API with the daemon

            async def start_daemon_and_polling_loops():
                polling_coro = api.poll()  # pylint: disable=assignment-from-no-return
                # Create task from polling coroutine and schedule for execution
                # Note: Unlike `asyncio.create_task`, `loop.create_task` works in Python < 3.7
                polling_task = loop.create_task(polling_coro)  # type: asyncio.Task

                # Run the blocking Pyro daemon request loop in a dedicated thread and `await` until finished
                # (`terminate_daemon` event set)
                with concurrent.futures.ThreadPoolExecutor() as pool:
                    try:
                        loop_daemon_until_event_set = partial(daemon.requestLoop,
                                                              lambda: not self.terminate_daemon.is_set())
                        await loop.run_in_executor(pool, loop_daemon_until_event_set)
                    finally:
                        LOGGER.debug("Canceling polling task.")
                        polling_task.cancel()

            loop = asyncio.get_event_loop()
            try:
                # Run event loop until request loops finished
                loop.run_until_complete(start_daemon_and_polling_loops())
            finally:
                loop.close()
            LOGGER.debug("Exiting service.")
            time.sleep(2.0)  # Allows data scraping

        return

    def start(self, mp_ctx, cloud_client_serialized: str) -> str:
        """
        Runs the scheduler as a Pyro4 object on a predetermined location in a subprocess.
        :param mp_ctx: Multiprocessing context, e.g. `multiprocessing.get_context("spawn")`
        :param cloud_client_serialized: Dill-serialized CloudClient instance
        :return: Pyro URI
        """

        if self.is_running():
            raise RuntimeError("Running already at {uri}".format(uri=self.uri))
        LOGGER.info("Starting service...")
        proc = mp_ctx.Process(target=self.daemonize, args=[cloud_client_serialized])
        proc.daemon = True
        proc.start()
        proc.join()
        time.sleep(DAEMON_BOOT_WAIT_TIME)  # Allow Pyro to boot up
        LOGGER.info("Service started.")
        return self.uri

    def stop(self) -> bool:
        if self.is_running():
            if not self.terminate_daemon:
                raise RuntimeError("Terminate daemon event does not exist. "
                                   "The stop() method may have called from the wrong process.")
            self.terminate_daemon.set()  # Flag for requestLoop to terminate
            with Pyro4.Proxy(self.uri) as pyro_proxy:
                # triggers checking loopCondition
                pyro_proxy._pyroBind()  # pylint: disable=protected-access
            return True
        return False



"""Module to define different types used through Meeshkan codebase"""
from typing import NewType, Dict, Any, List, Tuple
import time

class ScalarIndexPairing:
    """Represents a coupling between a scalar value and an index. If no index is provided, time.monotonic() is used
    to maintain consistent scale."""
    def __init__(self, value: float, idx=None):
        self.value = value
        self.idx = idx if idx is not None else time.monotonic()

Token = NewType("Token", str)
Payload = Dict[str, Any]
# Keys are scalar names, value is a list of tuples, indicating wall time and value for that scalar
HistoryByScalar = Dict[str, List[ScalarIndexPairing]]



import os
from ..core.service import Service
from ..exceptions import JobNotFoundException

__all__ = ["report_scalar"]


def report_scalar(val_name, value, *vals) -> bool:
    """Reports scalars to the meeshkan service API

    :param val_name: The name of the scalar to report
    :param value: The value of the scalar
    :param vals: any additional value_name, value to add.
    :param condition: A callable that accepts all scalars registered in this statement, and return True or False,
        indicating whether a notification is required.
    """
    if len(vals) % 2:  # Invalid number of additional scalar arguments given
        raise RuntimeError("Invalid number of arguments given - did you forget a name/value?")

    pid = os.getpid()
    with Service().api as proxy:
        try:
            proxy.report_scalar(pid, val_name, value)
            for name, val in zip(vals[::2], vals[1::2]):
                proxy.report_scalar(pid, name, val)
        except JobNotFoundException:
            return False
    return True



import logging
import queue
import threading
from typing import Any, Dict, List, Tuple, Callable, Optional, Union  # For self-documenting typing
from pathlib import Path
import uuid
import os
import asyncio

from .tracker import TrackingPoller, TrackerBase
from .job import ProcessExecutable, JobStatus, Job, SageMakerJob
from .config import JOBS_DIR
from ..exceptions import JobNotFoundException
from ..notifications.notifiers import Notifier

# Do not expose anything by default (internal module)
__all__ = []  # type: List[str]


LOGGER = logging.getLogger(__name__)


class QueueProcessor:
    """
    Process items from queue in a new thread. Usage:
    1) queue_processor.start(queue_=..., process_item=...)
    2) queue_processor.schedule_stop()
    3) queue_processor.wait_stop()
    The two shutdown steps ensure that the caller can cancel the currently running task between them,
    ensuring that processor exits without processing new jobs.
    """
    def __init__(self):
        self._stop_event = threading.Event()
        self._queue = None
        self._thread = None

    def start(self, queue_, process_item):
        """
        Read and handle tasks from queue `queue_` until (1) queue item is None or (2) stop_event is set. Note that
        the currently running job is not forced to cancel: that should be done from another thread, letting queue reader
        to check loop condition. `QueueProcessor` is NOT safe for reuse for processing other queues.
        :param queue_: Synchronized queue
        :param process_item: Callback called with queue item as argument
        :return:
            """
        self._queue = queue_
        self._thread = threading.Thread(target=self.__process, args=(process_item,))
        self._stop_event.clear()
        self._thread.start()

    def __process(self, process_item):
        if not self.is_running():
            raise RuntimeError("QueueProcessor must be started first.")
        while not self._stop_event.is_set():
            item = self._queue.get(block=True)
            if item is None or self._stop_event.is_set():
                break
            process_item(item)
            self._queue.task_done()

    def schedule_stop(self):
        if not self.is_running():
            return
        self._stop_event.set()  # Signal exit to worker thread, required as "None" may not be next task
        self._queue.put(None, block=False)  # Signal exit if thread is blocking

    def is_running(self):
        return self._thread is not None and self._thread.is_alive()

    def wait_stop(self):
        if self.is_running():
            self._thread.join()
            self._thread = None
            self._queue = None


class Scheduler:
    def __init__(self, queue_processor: QueueProcessor, notifier: Notifier = None):
        self._queue_processor = queue_processor
        self.submitted_jobs = dict()  # type: Dict[uuid.UUID, Job]
        self._job_queue = queue.Queue()  # type: queue.Queue
        self._running_job = None  # type: Optional[Job]
        self._job_poller = TrackingPoller(self.__query_and_report)
        self._event_loop = asyncio.get_event_loop()  # Save the event loop for out-of-thread operations
        self._notifier = notifier  # type: Optional[Notifier]

    # Properties and Python magic

    @property
    def jobs(self):  # Needed to access internal list of jobs as object parameters are unexposable, only methods
        return list(self.submitted_jobs.values())

    @property
    def is_running(self):
        return self._queue_processor.is_running()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    # Job handling methods
    def _handle_job(self, job: Job) -> None:
        LOGGER.debug("Handling job: %s", job)
        if job.status.stale:
            return
        self._running_job = job

        task = None  # type: Optional[asyncio.Task]
        if job.poll_time:
            # Create and schedule a task from the Polling job, so we can cancel it without killing the event loop
            task = self._event_loop.create_task(self._job_poller.poll(job.id, job.poll_time))
        if self._notifier:
            self._notifier.notify_job_start(job)
        try:
            job.launch_and_wait()
        except Exception:  # pylint:disable=broad-except
            LOGGER.exception("Running job failed")
        finally:
            if task:
                task.cancel()

        if self._notifier:
            self._notifier.notify_job_end(job)
        self._running_job = None
        LOGGER.debug("Finished handling job: %s", job)

    def submit_job(self, job: Job):
        job.status = JobStatus.QUEUED
        self.submitted_jobs[job.id] = job
        self._job_queue.put(job)  # TODO Blocks if queue full
        LOGGER.debug("Job submitted: %s", job)

    def stop_job(self, job_id: uuid.UUID):
        if job_id not in self.submitted_jobs:
            LOGGER.debug("Ignoring stopping unknown job with ID: %s", str(job_id))
            return
        self.submitted_jobs[job_id].cancel()

    # Tracking methods

    def __get_job_by_pid(self, pid):
        job_id = [job.id for job in self.jobs if job.pid == pid]
        if not job_id:
            raise JobNotFoundException(job_id=str(pid))
        return job_id[0]

    def add_condition(self, pid: int, *vals: str, condition: Callable[[float], bool], only_relevant: bool):
        """Adds a new condition for a job that matches the given process id.
        :param pid: process id
        :param vals: list of scalar names (strings)
        :param condition: a callable that accepts as many values as vals, and returns a boolean whether a condition has
            been met
        :param only_relevant: a boolean, whether or not only the values relevant to the condition should be plotted when
            this condition is met
        """
        job_id = self.__get_job_by_pid(pid)
        self.submitted_jobs[job_id].add_condition(*vals, condition=condition, only_relevant=only_relevant)

    def report_scalar(self, pid, name, val):
        # Find the right job id
        job_id = self.__get_job_by_pid(pid)
        condition = self.submitted_jobs[job_id].add_scalar_to_history(scalar_name=name, scalar_value=val)
        if condition and self._notifier:
            # TODO - we can add the condition that triggered the notification...
            names = condition.names if condition.only_relevant else list()
            vals, imgpath = self.query_scalars(*names, job_id=job_id, latest_only=False, plot=True)
            self._notifier.notify(self.submitted_jobs[job_id], imgpath, n_iterations=-1)
            if imgpath is not None:
                os.remove(imgpath)

    def query_scalars(self, *names: Tuple[str, ...], job_id, latest_only: bool = True, plot: bool = False):
        job = self.submitted_jobs.get(job_id)
        if not job:
            raise JobNotFoundException(job_id=str(job_id))
        return job.get_updates(*names, plot=plot, latest=latest_only)

    def __query_and_report(self, job_id: uuid.UUID):
        if self._notifier:
            # Get updates; TODO - vals should be reported once we update schema...
            vals, imgpath = self.query_scalars(job_id=job_id, latest_only=True, plot=True)
            if vals:  # Only send updates if there exists any updates
                self._notifier.notify(self.submitted_jobs[job_id], imgpath, n_iterations=-1)
            if imgpath is not None:
                os.remove(imgpath)

    # Scheduler service methods

    def start(self):
        if not self._queue_processor.is_running():
            LOGGER.debug("Start queue processor")
            self._queue_processor.start(queue_=self._job_queue, process_item=self._handle_job)
            LOGGER.debug("Queue processor started")

    def stop(self):
        # self._job_poller.stop()
        self._queue_processor.schedule_stop()
        if self._running_job is not None:
            # TODO Add an option to not cancel the currently running job?
            self._running_job.cancel()
        if self._queue_processor.is_running():
            # Wait for the thread to finish
            self._queue_processor.wait_stop()



"""Watch a running SageMaker job."""
import asyncio
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Tuple
import logging
import os
import threading

import pandas as pd

from .job import JobStatus, SageMakerJob, BaseJob
from ..exceptions import SageMakerNotAvailableException, JobNotFoundException, DeferredImportException

try:
    # Just in case we make boto3 dependency optional
    import boto3
except ImportError as ex:
    boto3 = DeferredImportException(ex)

try:
    # Sagemaker Python SDK is optional
    import sagemaker
except ImportError as ex:
    sagemaker = DeferredImportException(ex)


LOGGER = logging.getLogger(__name__)

# Do not expose anything by default (internal module)
__all__ = []  # type: List[str]


class SageMakerHelper:
    SAGEMAKER_STATUS_TO_JOB_STATUS = {
        "InProgress": JobStatus.RUNNING,
        "Completed": JobStatus.FINISHED,
        "Failed": JobStatus.FAILED,
        "Stopping": JobStatus.RUNNING,  # TODO Create status for this?
        "Stopped": JobStatus.CANCELLED_BY_USER
    }

    def __init__(self, client=None, sagemaker_session=None):
        """
        Init SageMaker helper in the disabled state.
        :param client: SageMaker client built with boto3.client("sagemaker") used for low-level connections to SM API
        :param sagemaker_session: SageMaker Python SDK session
        """
        self.client = client
        self.connection_tried = False
        self.connection_succeeded = False
        self._error_message = None  # type: Optional[str]
        self.sagemaker_session = sagemaker_session
        self.lock = threading.Lock()
        self.analytics_by_job_name = {}  # type: Dict[str, sagemaker.analytics.TrainingJobAnalytics]

    @property
    def __has_client(self):
        return self.client is not None

    def check_or_build_connection(self):
        """
        Check that SageMaker connection exists. Tries to create if not yet attempted.
        ALWAYS call this before attempting to access SageMaker.
        :raises SageMakerNotAvailableException: If connection to SageMaker cannot be established.
        :return:
        """
        if self.connection_tried:
            if self.connection_succeeded:
                return
            if not self.connection_succeeded:
                raise SageMakerNotAvailableException(message=self._error_message or None)

        self.connection_tried = True

        self.client = self.client or SageMakerHelper.build_client_or_none()

        if not self.client:
            self._error_message = "Could not create boto client. Check your credentials"
            raise SageMakerNotAvailableException(self._error_message)

        self.sagemaker_session = self.sagemaker_session or sagemaker.session.Session(sagemaker_client=self.client)

        try:
            self.client.list_training_jobs()
            LOGGER.info("SageMaker client successfully verified.")
        except Exception:  # pylint:disable=broad-except
            LOGGER.exception("Could not verify SageMaker connection")
            self._error_message = "Could not connect to SageMaker. Check your authorization."
            raise SageMakerNotAvailableException(self._error_message)

        self.connection_succeeded = True

    @staticmethod
    def build_client_or_none():
        """
        :return: SageMaker boto3 client or None if failed
        """
        try:
            return boto3.client("sagemaker")
        except Exception:  # pylint: disable=broad-except
            return None

    def get_job_status(self, job_name) -> JobStatus:
        """
        Get job status from SageMaker API. Use this to start monitoring jobs and to check they exist.
        :param job_name: Name of the SageMaker training job
        :raises SageMakerNotAvailableException:
        :raises JobNotFoundException: If job was not found.
        :return: Job status
        """
        with self.lock:
            self.check_or_build_connection()

        try:
            training_job = self.client.describe_training_job(TrainingJobName=job_name)
        except self.client.exceptions.ClientError:
            raise JobNotFoundException

        status = training_job['TrainingJobStatus']
        return SageMakerHelper.SAGEMAKER_STATUS_TO_JOB_STATUS[status]

    def wait_for_job_finish(self, job_name: str) -> JobStatus:
        """
        Wait for SageMaker job to finish (blocking).
        :param job_name: SageMaker training job name
        :raises Exception: If job does not finish cleanly (is stopped, for example) or waiting took too long
        :return JobStatus: Job status after waiting
        """
        with self.lock:
            self.check_or_build_connection()
        LOGGER.info("Started waiting for job %s to finish.", job_name)
        waiter = self.client.get_waiter('training_job_completed_or_stopped')
        attempt_delay_secs = 60
        max_attempts = 60 * 24 * 3  # Three days
        waiter.wait(
            TrainingJobName=job_name,
            WaiterConfig={
                'Delay': attempt_delay_secs,
                'MaxAttempts': max_attempts  # TODO Wait longer?
            }
        )
        job_status = self.get_job_status(job_name=job_name)
        LOGGER.info("Job %s finished with status %s", job_name, job_status)

        if not job_status.is_processed and not job_status.stale:
            waited_hours = max_attempts * attempt_delay_secs / 3600
            LOGGER.exception("Exited waiter after waiting %f hours", waited_hours)
            raise RuntimeError("Did not expect to wait for more than {hours} hours".format(hours=waited_hours))
        return job_status

    def get_training_job_analytics_df(self, job_name: str):
        with self.lock:
            self.check_or_build_connection()

        LOGGER.debug("Checking for updates for job %s", job_name)
        analytics = self.analytics_by_job_name.setdefault(job_name,
                                                          sagemaker.analytics.TrainingJobAnalytics(
                                                              training_job_name=job_name,
                                                              sagemaker_session=self.sagemaker_session))
        return analytics.dataframe(force_refresh=True)


class JobScalarHelper:

    def __init__(self, job: BaseJob):
        self.job = job
        self.last_timestamp_by_metric = {}  # type: Dict[str, float]

    def add_new_scalars_from(self, metrics_dataframe: pd.DataFrame) -> bool:
        """
        Add all new records from `metrics_dataframe` to job scalar history, keeping track of the previously
        seen maximum timestamp. It is assumed that for a given metric, all new records have timestamps larger than
        the previously seen maximum timestamp.
        :param metrics_dataframe: DataFrame with records with names "metric_name", "value", "timestamp"
        :return: Boolean denoting if new values were added to job scalar history
        """
        metrics_grouped_by_name = metrics_dataframe.groupby(by="metric_name")

        added_new_metrics = False

        for metric_name, metrics_for_name in metrics_grouped_by_name:
            max_timestamp_for_name = metrics_for_name["timestamp"].max()
            previous_last_timestamp_or_none = self.last_timestamp_by_metric.get(metric_name, float("-inf"))

            new_records_df = metrics_for_name[metrics_for_name.timestamp > previous_last_timestamp_or_none]

            for _, record in new_records_df.iterrows():
                added_new_metrics = True
                self.job.add_scalar_to_history(scalar_name=record['metric_name'], scalar_value=record['value'])

            self.last_timestamp_by_metric[metric_name] = max_timestamp_for_name
        return added_new_metrics


class SageMakerJobMonitor:
    MINIMUM_POLLING_INTERVAL_SECS = 60

    def __init__(self,
                 event_loop=None,
                 sagemaker_helper: Optional[SageMakerHelper] = None,
                 notify_start: Optional[Callable[[BaseJob], Any]] = None,
                 notify_update: Optional[Callable[[BaseJob, str, int, Optional[str]], Any]] = None,
                 notify_finish: Optional[Callable[[BaseJob], Any]] = None,
                 scalar_helper_factory: Optional[Callable[[BaseJob], JobScalarHelper]] = None):
        super().__init__()
        # self._notify = notify_function
        self._event_loop = event_loop or asyncio.get_event_loop()
        self.sagemaker_helper = sagemaker_helper or SageMakerHelper()  # type: SageMakerHelper
        self.notify_start = notify_start
        self.notify_finish = notify_finish
        self.notify_update = notify_update
        self.job_scalar_helper_factory = scalar_helper_factory or JobScalarHelper

    def start(self, job: SageMakerJob) -> asyncio.Task:
        self.sagemaker_helper.check_or_build_connection()
        return self._event_loop.create_task(self.monitor(job))

    async def monitor(self, job: SageMakerJob):
        update_polling_task = self._event_loop.create_task(self.poll_updates(job))  # type: asyncio.Task
        wait_for_finish_future = \
            self._event_loop.run_in_executor(
                None, self.sagemaker_helper.wait_for_job_finish, job.name)
        try:
            job_status = await wait_for_finish_future
        except Exception as ex:  # pylint:disable=broad-except
            if isinstance(ex, asyncio.CancelledError):
                raise ex
            LOGGER.exception("Failed waiting for job to finish")
            job_status = self.sagemaker_helper.get_job_status(job_name=job)
        finally:
            if not update_polling_task.done():
                LOGGER.info("Canceling polling for job %s", job.name)
                try:
                    update_polling_task.cancel()
                except Exception:  # pylint:disable=broad-except
                    LOGGER.exception("Canceling the task failed")

        job.status = job_status
        if self.notify_finish:
            LOGGER.info("Notifying finish for job %s with status %s", job.name, job.status)
            self.notify_finish(job)

    async def poll_updates(self, job: BaseJob):
        if not isinstance(job, SageMakerJob):
            raise RuntimeError("SageMakerJobMonitor can only monitor SageMakerJobs.")

        if job.status.is_processed:
            LOGGER.info("SageMaker job %s already finished, returning", job.name)
            return

        sleep_time = max(job.poll_time, SageMakerJobMonitor.MINIMUM_POLLING_INTERVAL_SECS)
        LOGGER.debug("Starting SageMaker job tracking for job %s with polling interval of %f seconds.",
                     job.name, sleep_time)

        if job.status.is_launched and self.notify_start:
            self.notify_start(job)

        job_scalar_helper = self.job_scalar_helper_factory(job)

        try:
            while True:
                await self.check_and_apply_updates(job=job, job_scalar_helper=job_scalar_helper)

                if job.status.is_processed:
                    break

                await asyncio.sleep(sleep_time)

            LOGGER.info("Stopped monitoring SageMakerJob %s, got status %s", job.name, job.status)
        except asyncio.CancelledError:
            LOGGER.debug("SageMakerJob tracking cancelled for job %s", job.name)
        except Exception:  # pylint:disable=broad-except
            LOGGER.exception("Polling for updates failed")
            # Ignore

    async def check_and_apply_updates(self, job: BaseJob, job_scalar_helper: JobScalarHelper):
        LOGGER.debug("Checking updates for job %s", job.name)
        previous_status = job.status
        job.status = await self._event_loop.run_in_executor(None, self.sagemaker_helper.get_job_status, job.name)
        LOGGER.debug("Job %s: previous status %s, current status %s", job.name, previous_status, job.status)
        if not previous_status.is_launched and job.status.is_launched and self.notify_start:
            self.notify_start(job)

        added_new_scalars = False

        try:
            get_training_job_analytics = partial(self.sagemaker_helper.get_training_job_analytics_df, job_name=job.name)
            metrics_df = await self._event_loop.run_in_executor(None, get_training_job_analytics)
            if not metrics_df.empty:
                added_new_scalars = job_scalar_helper.add_new_scalars_from(metrics_df)
        except Exception as ex:  # pylint:disable=broad-except
            # Reading TrainingJobAnalytics routinely throws an exception, handle it here
            # TODO Only catch a more specific exception to avoid getting into failure loop?
            if isinstance(ex, asyncio.CancelledError):
                raise ex
            LOGGER.exception("Checking for SageMaker job metrics failed, ignoring.")

        if added_new_scalars:
            # Something new to report
            self._event_loop.run_in_executor(None, self.__query_and_report, job)

    # TODO Remove duplicate code with Scheduler
    def query_scalars(self, *names: Tuple[str, ...], job, latest_only: bool = True, plot: bool = False):
        if not job:
            raise JobNotFoundException(job_id=str(job.id))
        return job.get_updates(*names, plot=plot, latest=latest_only)

    def __query_and_report(self, job):
        if self.notify_update:
            # Get updates; TODO - vals should be reported once we update schema...
            vals, imgpath = self.query_scalars(job=job, latest_only=True, plot=True)
            if vals:  # Only send updates if there exists any updates
                self.notify_update(job, imgpath, n_iterations=-1)
            if imgpath is not None:
                os.remove(imgpath)

    def create_job(self, job_name: str, poll_interval: Optional[float] = None) -> SageMakerJob:
        sagemaker_helper = self.sagemaker_helper
        status = sagemaker_helper.get_job_status(job_name)
        return SageMakerJob(job_name=job_name,
                            status=status,
                            poll_interval=poll_interval)



"""
Build the whole dependency chain leading to Api instance exposed by Pyro.
"""

__all__ = []  # type: ignore


def _build_api(service, cloud_client):
    from meeshkan.core.api import Api
    from meeshkan.notifications.notifiers import CloudNotifier, LoggingNotifier, NotifierCollection
    from meeshkan.core.tasks import TaskPoller
    from meeshkan.core.scheduler import Scheduler, QueueProcessor
    from meeshkan.core.config import ensure_base_dirs as ensure_base_dirs_
    from meeshkan.core.logger import setup_logging as setup_logging_
    from meeshkan.core.sagemaker_monitor import SageMakerJobMonitor

    ensure_base_dirs_()
    setup_logging_(silent=True)

    cloud_notifier = CloudNotifier(name="Cloud Service", post_payload=cloud_client.post_payload,
                                   upload_file=cloud_client.post_payload_with_file)
    logging_notifier = LoggingNotifier(name="Local Service")

    task_poller = TaskPoller(cloud_client.pop_tasks)
    queue_processor = QueueProcessor()

    notifier_collection = NotifierCollection(*[cloud_notifier, logging_notifier])

    scheduler = Scheduler(queue_processor=queue_processor, notifier=notifier_collection)

    sagemaker_job_monitor = SageMakerJobMonitor(notify_start=notifier_collection.notify_job_start,
                                                notify_update=notifier_collection.notify,
                                                notify_finish=notifier_collection.notify_job_end)

    api = Api(scheduler=scheduler,
              service=service,
              task_poller=task_poller,
              notifier=notifier_collection,
              sagemaker_job_monitor=sagemaker_job_monitor)
    api.add_stop_callback(cloud_client.close)
    return api



from .__version__ import __version__  # Conform to PEP-0396

from . import core
from .core import *  # pylint: disable=wildcard-import
from . import api
from .api import *  # pylint: disable=wildcard-import
from . import agent
from .agent import *  # pylint: disable=wildcard-import
from . import notifications
from .notifications import *  # pylint: disable=wildcard-import

from . import exceptions
from . import sagemaker
from .__utils__ import save_token


# Only make the following available by default
__all__ = ["__version__", "exceptions", "config", "save_token"]
__all__ += api.__all__
__all__ += agent.__all__
__all__ += core.__all__
__all__ += notifications.__all__

# Clean-up (make `meeshkan.core`, etc unavailable)
del core
del api
del agent
del notifications



from typing import Optional, List

import Pyro4

from ..core.service import Service
from ..core.job import SageMakerJob

__all__ = ["monitor"]  # type: List[str]

Pyro4.config.SERIALIZER = 'dill'


def monitor(job_name: str, poll_interval: Optional[float] = None):
    """
    Start monitoring a SageMaker training job.

    :param job_name: SageMaker training job name
    :param poll_interval: Polling interval in seconds, optional. Defaults to one hour.
    :return: SageMakerJob instance
    """
    with Service().api as proxy:
        sagemaker_job = proxy.monitor_sagemaker(job_name=job_name, poll_interval=poll_interval)  # type: SageMakerJob
        if sagemaker_job.status.is_processed:
            print("Job {job_name} is already finished with status {status}.".format(job_name=sagemaker_job.name,
                                                                                    status=sagemaker_job.status.name))
        else:
            print("Started monitoring job {job_name}, "
                  "currently in phase {status}".format(job_name=sagemaker_job.name, status=sagemaker_job.status.name))
        return sagemaker_job



from setuptools import find_packages, setup, Command
import stat
import os
from shutil import rmtree
import sys

# Package meta-data.
NAME = 'meeshkan'
DESCRIPTION = 'The Meeshkan Client for interactive machine learning'
URL = 'https://www.meeshkan.com/'
EMAIL = 'dev@meeshkan.com'
AUTHOR = 'Meeshkan Dev Team'
REQUIRES_PYTHON = '>=3.6.0'
SRC_DIR = 'meeshkan'  # Relative location wrt setup.py

# Required packages.
# Older version of requests because >= 2.21 conflicts with sagemaker.
REQUIRED = ['boto3', 'dill', 'requests<2.21', 'Click', 'pandas', 'Pyro4', 'PyYAML', 'tabulate', 'matplotlib']

DEV = ['jupyter', 'nbdime', 'pylint', 'pytest==4.0.2', 'pytest-cov', 'mypy', 'pytest-asyncio', 'sagemaker', 'sphinx',
       'sphinx-click', 'sphinx_rtd_theme']

# Optional packages
EXTRAS = {'dev': DEV,
          'devTF': DEV + ['tensorflow', 'tensorboard', 'keras'],
          'devTorch':  DEV + ['torch']}

# Entry point for CLI (relative to setup.py)
ENTRY_POINTS = ['meeshkan = meeshkan.__main__:cli']

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = '\n' + f.read()

# Load the package's __version__.py module as a dictionary.
about = dict()
with open(os.path.join(here, SRC_DIR, '__version__.py')) as f:
    exec(f.read(), about)


class SetupCommand(Command):
    """Base class for setup.py commands with no arguments"""
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))


class BuildDistCommand(SetupCommand):
    """Support setup.py upload."""
    description = "Build the package."

    def run(self):
        try:
            self.status("Removing previous builds...")
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution...")
        os.system("{executable} setup.py sdist bdist_wheel --universal".format(executable=sys.executable))
        sys.exit()


def build_docs():
    os.chdir("docs")
    os.system("sphinx-apidoc -f -e -o source/ ../meeshkan/")
    os.system("sphinx-build -M html -D version={version} source build".format(version=about['__version__']))


class BuildDocumentationCommand(SetupCommand):
    """Builds the sphinx documentation"""
    description = "Builds the sphinx documentation."

    def run(self):
        try:
            self.status("Removing previous builds...")
            rmtree(os.path.join(here, 'docs/build'))
        except OSError:
            pass

        self.status("Building documentation...")
        build_docs()
        self.status("Docs were built. Now change to fresh branch, `git add docs/build`, `git commit -a` and do "
                    "`git subtree push --prefix docs/build origin gh-pages` to push the build to `gh-pages` branch.")
        sys.exit()


class UploadCommand(SetupCommand):
    """Support setup.py upload."""
    description = "Build and publish the package."

    def run(self):
        try:
            self.status("Removing previous builds...")
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution...")
        os.system("{executable} setup.py sdist bdist_wheel --universal".format(executable=sys.executable))

        self.status("Uploading the package to PyPI via Twine...")
        os.system("twine upload dist/*")

        self.status("Pushing git tags...")
        os.system("git tag v{about}".format(about=about['__version__']))
        os.system("git push --tags")

        sys.exit()


class TestCommand(SetupCommand):
    """Support setup.py test."""
    description = "Run local test if they exist"

    def run(self):
        os.system("pytest")
        sys.exit()


setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=('tests',)),
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    license='Apache 2.0',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Operating System :: MacOS',
        'Operating System :: POSIX',
        'Operating System :: Unix'
    ],
    entry_points={'console_scripts': ENTRY_POINTS},
    cmdclass={'dist': BuildDistCommand, 'upload': UploadCommand, 'test': TestCommand,
              'doc': BuildDocumentationCommand}
)



"""
Code related to tasks invoked by the cloud.
"""
import asyncio
from enum import Enum
import logging
from typing import Callable, List
from uuid import UUID

LOGGER = logging.getLogger(__name__)

# Do not expose anything by default (internal module)
__all__ = []  # type: List[str]


class TaskType(Enum):
    StopJobTask = 0


class Task:
    def __init__(self, job_id: UUID, task_type: TaskType):
        self.job_id = job_id
        self.type = task_type


class TaskPoller:
    def __init__(self, pop_tasks: Callable[[], List[Task]]):
        """
        Polls new tasks from the server.
        :param pop_tasks: Asynchronous method for fetching new tasks
        """
        self._pop_tasks = pop_tasks

    async def poll(self, handle_task, delay=10):
        """
        Polling for tasks.
        :param handle_task: Async task handler. Must NOT block the event loop.
        :param delay: Time in seconds to wait between requesting new tasks. Should be reasonably long to avoid
        bombarding the server.
        :return:
        """
        loop = asyncio.get_event_loop()
        try:
            while True:
                try:
                    tasks = await loop.run_in_executor(None, self._pop_tasks)  # type: List[Task]
                    for task in tasks:
                        await handle_task(task)
                except Exception as ex:  # pylint:disable=broad-except
                    if isinstance(ex, asyncio.CancelledError):
                        raise
                    LOGGER.exception("Failed fetching or handling new tasks.")
                await asyncio.sleep(delay)
        except asyncio.CancelledError:
            LOGGER.debug("Polling canceled.")
            raise



import time
from typing import Tuple
from enum import Enum, auto


class NotificationStatus(Enum):
    SUCCESS = auto()
    FAILED = auto()


class NotificationType(Enum):
    JOB_START = auto()
    JOB_END = auto()
    JOB_UPDATE = auto()


class NotificationWithStatusTime:
    """Class to hold a notification type, timestamp and status"""
    def __init__(self, notification_type: NotificationType, status: NotificationStatus):
        self.time = time.strftime(time.strftime("%D %H:%M:%S"))
        self.type = notification_type
        self.status = status



import os
import logging
from pathlib import Path
from typing import Optional, List

LOGGER = logging.getLogger(__name__)

PACKAGE_PATH = Path(os.path.dirname(__file__))  # type: Path

CONFIG_PATH = PACKAGE_PATH.joinpath('config.yaml')
LOG_CONFIG_FILE = PACKAGE_PATH.joinpath('logging.yaml')

BASE_DIR = Path.home().joinpath('.meeshkan')
JOBS_DIR = BASE_DIR.joinpath('jobs')
LOGS_DIR = BASE_DIR.joinpath('logs')

CREDENTIALS_FILE = BASE_DIR.joinpath('credentials')

CONFIG = None  # type: Optional[Configuration]
CREDENTIALS = None  # type: Optional[Credentials]


# Don't automatically expose anything to top level, as the entire module is loaded as-is
__all__ = []  # type: List[str]


del logging  # Clean-up (only leaves Path available in this module)
del os
del List
del Optional


def ensure_base_dirs(verbose=True):

    def create_dir_if_not_exist(path: Path):
        if not path.is_dir():
            # Print instead of logging as loggers may not have been configured yet
            if verbose:
                print("Creating directory {path}".format(path=path))
            path.mkdir()

    create_dir_if_not_exist(BASE_DIR)
    create_dir_if_not_exist(JOBS_DIR)
    create_dir_if_not_exist(LOGS_DIR)


class Configuration:
    def __init__(self, cloud_url):
        self.cloud_url = cloud_url

    @staticmethod
    def from_yaml(path: Path = CONFIG_PATH):
        import yaml

        LOGGER.debug("Reading configuration from %s", path)
        if not path.is_file():
            raise FileNotFoundError("File {path} not found".format(path=path))
        with path.open('r') as file:
            config = yaml.safe_load(file.read())
        return Configuration(cloud_url=config['cloud']['url'])


class Credentials:

    def __init__(self, refresh_token):
        self.refresh_token = refresh_token

    @staticmethod
    def from_isi(path: Path = CREDENTIALS_FILE):
        import configparser

        LOGGER.debug("Reading credentials from %s", path)
        if not path.is_file():
            raise FileNotFoundError("Create file {path} first.".format(path=path))
        conf = configparser.ConfigParser()
        conf.read(str(path))
        return Credentials(refresh_token=conf['meeshkan']['token'])

    @staticmethod
    def to_isi(refresh_token: str, path: Path = CREDENTIALS_FILE):
        """Creates the credential file with given refresh token. Overrides previous token if exists.
        Does not create missing folders along `path`, instead, raises FileNotFound exception.

        """
        with path.open("w") as credential_file:
            credential_file.write("[meeshkan]\ntoken={token}\n".format(token=refresh_token))
            credential_file.flush()


def init_config(config_path: Path = CONFIG_PATH, credentials_path: Path = CREDENTIALS_FILE, force_refresh=False):
    """Allows a one-time initialization of CONFIG and CREDENTIALS."""
    global CONFIG, CREDENTIALS  # pylint:disable=global-statement
    if CONFIG is None or force_refresh:
        CONFIG = Configuration.from_yaml(config_path)
    if CREDENTIALS is None or force_refresh:
        CREDENTIALS = Credentials.from_isi(credentials_path)
    return CONFIG, CREDENTIALS



# Ignore mypy tests for this file; Attributes for the `meeshkan` package are defined dynamically in
#     __init__.py, so mypy complains about attributes not existing (even though they're well defined).
#     examples for such errors: "error: Name 'meeshkan.config.Configuration' is not defined",
#                               "error: Module has no attribute "config"

import logging

from typing import Tuple

import meeshkan
from .core.cloud import CloudClient
from .core.service import Service
from .core.api import Api
from .exceptions import AgentNotAvailableException

__all__ = ["save_token"]

LOGGER = logging.getLogger(__name__)


def get_auth() -> Tuple[meeshkan.config.Configuration, meeshkan.config.Credentials]:  # type: ignore
    config, credentials = meeshkan.config.init_config()  # type: ignore
    return config, credentials


def save_token(token: str):
    meeshkan.config.ensure_base_dirs(verbose=False)  # type: ignore
    meeshkan.config.Credentials.to_isi(refresh_token=token)  # type: ignore


def _get_api() -> Api:
    service = Service()
    if not service.is_running():
        print("Start the service first.")
        raise AgentNotAvailableException()
    api = service.api  # type: Api
    return api


def _build_cloud_client(config: meeshkan.config.Configuration,  # type: ignore
                        credentials: meeshkan.config.Credentials) -> CloudClient:  # type: ignore
    cloud_client = CloudClient(cloud_url=config.cloud_url, refresh_token=credentials.refresh_token)
    return cloud_client



"""Module to enable remote querying of python processeses from outside the current process."""
import os
from typing import Union, List, Dict, Tuple, Optional, Callable, Any
from pathlib import Path
import time
import uuid
import tempfile
import logging
import asyncio
import inspect

from ..__types__ import HistoryByScalar, ScalarIndexPairing
from ..exceptions import TrackedScalarNotFoundException

TF_EXISTS = True
try:
    import tensorboard.backend.event_processing.event_accumulator as tfproto
except ModuleNotFoundError:
    TF_EXISTS = False  # Silently fail

LOGGER = logging.getLogger(__name__)


# Do not expose anything by default (internal module)
__all__ = []  # type: List[str]


class TrackingPoller:
    def __init__(self, notify_function: Callable[[uuid.UUID], Any]):
        self._notify = notify_function

    async def poll(self, job_id, poll_time: float):
        """Asynchronous polling function for scalars in given job"""
        LOGGER.debug("Starting job tracking for job %s", job_id)
        sleep_time = poll_time
        try:
            while True:
                await asyncio.sleep(sleep_time)  # Let other tasks run meanwhile
                self._notify(job_id)  # Synchronously notify of changes.
        except asyncio.CancelledError:
            LOGGER.debug("Job tracking cancelled for job %s", job_id)


class TrackerCondition:
    DEF_COOLDOWN_PERIOD = 30  # 30 seconds interval default cooldown period
    def __init__(self, *value_names: str, condition: Callable[[float], bool], title: str,
                 default_value=1, cooldown_period: int = None, only_relevant: bool = False):
        """
        Initializes a new condition for tracking
        :param value_names: List of scalar names (strings)
        :param condition: A callable that accepts as many values as the length of value_names
        :param title: An optional title for this condition to be reported with the notification
        :param default_value: A default value to use when scalar values are missing (default is 1)
        :param cooldown_period: A cooldown interval (in seconds) from last notification, to prevent spamming when the
            condition is met (default is 30 seconds)
        :param only_relevant: A boolean flag to represent whether when this condition is met, only values relevant to
            this condition should be reported (default is False, i.e. report all scalars for relevant job)
        """
        if len(value_names) != len(inspect.signature(condition).parameters):
            raise RuntimeError("Number of arguments for condition {func} does not"
                               "match given number of arguments {vals}!".format(func=condition, vals=value_names))
        self.names = value_names  # type: Tuple[str, ...]
        self.condition = condition
        self.title = title or str(condition)
        self.default = default_value
        self.cooldown_period = cooldown_period or TrackerCondition.DEF_COOLDOWN_PERIOD
        self.last_time_condition_met = 0.0  # The last time condition() returned True
        self.only_relevant = only_relevant

    def __contains__(self, val_name: str):
        return val_name in self.names

    def __len__(self):
        return len(self.names)

    def __call__(self, **kwargs: float) -> bool:
        # match args/kwargs to given names
        if kwargs:
            vals = [kwargs.get(name, self.default) for name in self.names]  # type: List[float]
        else:
            vals = [self.default] * len(self.names)

        condition_result = self.condition(*vals)
        has_enough_time_passed = (time.monotonic() - self.last_time_condition_met) >= self.cooldown_period
        result = condition_result and has_enough_time_passed
        if result:
            self.last_time_condition_met = time.monotonic()
        return result


class TrackerBase:
    DEF_IMG_EXT = "png"
    """Defines common API for Tracker objects"""
    def __init__(self):
        # History of tracked information, var_name: list(vals)
        self._history_by_scalar = dict()  # type: HistoryByScalar
        # Last index which was submitted to cloud, used for statistics
        self._last_index = dict()  # type: Dict[str, int]
        self._conditions = list()  # type: List[TrackerCondition]
        self._shared_index = 0

    def add_tracked(self, val_name: str, value: float) -> Optional[TrackerCondition]:
        # Add/create to dictionaries
        if val_name in self._history_by_scalar and self._history_by_scalar[val_name][-1].idx == self._shared_index:
            self._shared_index += 1
        value_with_index = ScalarIndexPairing(value, self._shared_index)
        self._history_by_scalar.setdefault(val_name, list()).append(value_with_index)
        self._last_index.setdefault(val_name, -1)  # Initial index
        # Verify with conditions
        for condition in self._conditions:
            if val_name in condition:  # Condition is relevant for this scalar
                existing_values = [name for name in condition.names if name in self._history_by_scalar]
                kwargs = {name: self._history_by_scalar[name][-1].value for name in existing_values}
                if condition(**kwargs):
                    return condition
        return None

    def add_condition(self, *vals, condition: Callable[[float], bool], title: str = "", default_value=1,
                      only_relevant: bool):
        """Adds a condition for this tracker. Once a condition is met, it is reported immediately.
        If a variable listed in vals does not exist, the given default value (or 1 by default) will be sent instead. """
        self._conditions.append(TrackerCondition(*vals, condition=condition, title=title, default_value=default_value,
                                                 only_relevant=only_relevant))



    @staticmethod
    def generate_image(history: HistoryByScalar, output_path: Union[str, Path], title=None) -> Optional[str]:
        """
        Generates a plot from internal history to output_path

        If history contains no data, the image will not be generated.

        :return Absolute path to generated image if the image was generated, otherwise null
        """
        # Import matplotlib (or other future libraries) inside the function to prevent non-declaration in forked process
        import matplotlib
        matplotlib.use('svg')
        import matplotlib.pyplot as plt

        plt.clf()  # Clear figure
        for tag, vals in history.items():  # All all scalar values to plot
            if vals:  # Some values exist
                index_axis = list()
                value_axis = list()
                for value_with_index in vals:  # Separate axes
                    index_axis.append(value_with_index.idx)
                    value_axis.append(value_with_index.value)
                if len(vals) > 1:
                    plt.plot(index_axis, value_axis, label=tag, linewidth=1)
                else:  # scatter=plot the single point on x=0 :shrug:
                    plt.scatter(index_axis, value_axis, label=tag)

        if plt.gcf().axes:  # Would be empty if nothing was plotted
            plt.legend(loc='upper right')
            if title is not None:  # Title if given
                plt.title(title)
            fname, ext = os.path.splitext(output_path)  # Default extension if not provided
            if not ext:
                fname = os.path.abspath("{}.{}".format(fname, TrackerBase.DEF_IMG_EXT))
            plt.savefig(fname)
            return fname
        return None


    def _update_access(self, *names: str):
        if names:
            for name in names:
                self._last_index[name] = len(self._history_by_scalar[name]) - 1
        else:
            for val_name, vals, in self._history_by_scalar.items():
                self._last_index[val_name] = len(vals) - 1


    def _verify_scalar_names_exist(self, *names: str):
        for name in names:  # Verify all names are valid
            if name not in self._history_by_scalar:
                raise TrackedScalarNotFoundException(name=name)


    def get_updates(self, *names: str, plot: bool = True,
                    latest: bool = True) -> Tuple[HistoryByScalar, Optional[str]]:
        """Gets updates since last push update, possibly with an image

        :param names: names of value to lookup (or empty for all tracked history)
        :param plot: whether or not to plot the history and return the image path
        :param latest: whether or not to include all history, or just history since previous call
        :return tuple of data (HistoryByScalar) and location to image (if created, otherwise None)
        """
        if names:
            self._verify_scalar_names_exist(*names)
            data = dict()  # type: HistoryByScalar
            for value_name, values in self._history_by_scalar.items():
                for name in names:
                    if name == value_name:
                        data[name] = values
        else:
            data = dict(self._history_by_scalar)  # Create a copy

        imgname = None
        if plot:
            # pylint: disable=protected-access
            imgname = os.path.abspath(next(tempfile._get_candidate_names()))  # type: ignore
            imgname = self.generate_image(history=data, output_path=imgname)

        if latest:  # Trim data as needed
            for val_name, vals, in data.items():
                data[val_name] = vals[self._last_index[val_name] + 1:]

        self._update_access(*names)
        return data, imgname


    def refresh(self) -> None:
        """Cleans and updates"""
        self.clean()
        self.update()

    def get_statistics(self, val_name: str, from_beginning: bool = False):
        """Calculates viable statistics for further reporting for given value name"""
        pass

    def update(self) -> None:
        """Updates internal history stack for Tracker class"""
        pass

    def _clean(self) -> None:
        """Cleans internal history stack for Tracker class"""
        self._history_by_scalar = dict()
        self._last_index = dict()

    def clean(self) -> None:
        self._clean()


class TensorFlowTracker(TrackerBase):
    def __init__(self, path):
        global TF_EXISTS  # pylint: disable=global-statement
        if not TF_EXISTS:
            raise ModuleNotFoundError("Cannot instantiate a TensorFlowTracker without TensorFlow!")
        super().__init__()
        self.path = path
        self.ea_tracker = tfproto.EventAccumulator(path)
        self.update()

    def update(self):
        self.ea_tracker.Reload()
        for tag in self.ea_tracker.Tags()['scalars']:
            for scalar_event in self.ea_tracker.Scalars(tag):
                self.add_tracked(tag, scalar_event.value)

    def clean(self):
        self.ea_tracker._generator = tfproto._GeneratorFromPath(self.path)  # pylint: disable=protected-access
        self._clean()



import logging
import logging.config
from pathlib import Path
from typing import List

import yaml

from .config import LOG_CONFIG_FILE, LOGS_DIR

LOGGER = logging.getLogger(__name__)

# Do not expose anything by default (internal module)
__all__ = []  # type: List[str]


def setup_logging(log_config: Path = LOG_CONFIG_FILE, silent: bool = False):
    """Setup logging configuration
    This MUST be called before creating any loggers.
    """

    if not log_config.is_file():
        raise RuntimeError("Logging file {log_file} not found".format(log_file=log_config))

    with log_config.open() as log_file:
        config_orig = yaml.safe_load(log_file.read())

    def prepare_filenames(config):
        """
        Prepend `LOGS_DIR` to all 'filename' attributes listed for handlers in logging.yaml
        :param config: Configuration dictionary
        :return: Configuration with 'filename's prepended with LOGS_DIR
        """
        for handler_name in config['handlers'].keys():
            handler_config = config['handlers'][handler_name]
            if 'filename' in handler_config:
                filename = Path(handler_config['filename']).name
                handler_config['filename'] = str(LOGS_DIR.joinpath(filename))
        return config

    config = prepare_filenames(config_orig)
    logging.config.dictConfig(config)
    if silent:
        remove_non_file_handlers()


def remove_non_file_handlers():
    log = logging.getLogger()  # Root logger
    for handler in log.handlers.copy():
        if not isinstance(handler, logging.FileHandler):
            log.handlers.remove(handler)
    LOGGER.info("Deleted non-file handlers from logging")
